"""
This module contains helpers for the SUMM.AI API client
"""
import logging
import itertools

from html import unescape

from lxml.etree import SubElement, strip_tags
from lxml.html import fromstring, tostring

from django.conf import settings
from django.contrib import messages
from django.utils.translation import ugettext as _

from ..cms.constants import status
from ..cms.utils.translation_utils import ugettext_many_lazy as __

logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class TextField:
    """
    A class for simple text fields
    """

    #: The name of the corresponding model field
    #:
    #: :type: str
    name = ""
    #: The source text
    #:
    #: :type: str
    text = ""
    #: The translated text
    #:
    #: :type: str
    translated_text = ""
    #: The exception which occurred during translation, if any
    #:
    #: :type: Exception
    exception = None

    def __init__(self, name, translation):
        """
        Constructor initializes the class variables

        :param text: The text to be translated
        :type text: str
        """
        self.name = name
        self.text = getattr(translation, name, "")

    def translate(self, translated_text):
        """
        Translate the text of the current text field

        :param translated_text: The translated text
        :type translated_text: str
        """
        self.translated_text = translated_text

    def __repr__(self):
        """
        The representation used for logging

        :return: The canonical string representation of the text field
        :rtype: str
        """
        return f"<{type(self).__name__} (text: {self.text})>"


# pylint: disable=too-few-public-methods
class HTMLSegment(TextField):
    """
    A class for translatable HTML segments
    """

    #: The current HTML segment
    #:
    #: :type: lxml.html.HtmlElement
    segment = None

    # pylint: disable=super-init-not-called
    def __init__(self, segment):
        """
        Convert the lxml tree element to a flat text string.
        Preserve <br> tags as new lines characters.
        Remove all inner tags but keep their text content.
        Unescape all special HTML entities into unicode characters.

        :param segment: The current HTML segment
        :type segment: ~integreat_cms.cms.models.regions.region.Region
        """
        self.segment = segment
        # Preserve new line tags
        for br in self.segment.iter("br"):
            br.tail = "\n" + br.tail if br.tail else "\n"
        # Strip all inner tags
        strip_tags(self.segment, "*")
        # Unescape to convert umlauts etc. to unicode
        self.text = unescape(self.segment.text_content())

    def translate(self, translated_text):
        """
        Translate the current HTML segment and create new sub elements for line breaks

        :param translated_text: The translated text
        :type translated_text: str
        """
        # Split the text by newlines characters
        lines = translated_text.splitlines()
        # Take the first line as initial text
        self.segment.text = lines[0]
        # If there are more than one line returned, insert <br> tags
        for line in lines[1:]:
            SubElement(self.segment, "br").tail = line


# pylint: disable=too-few-public-methods
class HTMLField:
    """
    A class for more complex HTML fields which are splitted into segments
    """

    #: The name of the corresponding model field
    #:
    #: :type: str
    name = ""
    #: The list of HTML segments
    #:
    #: :type: list [ lxml.html.HtmlElement ]
    segments = []
    #: The current HTML stream
    #:
    #: :type: lxml.html.HtmlElement
    html = None

    def __init__(self, name, translation):
        """
        Parse the HTML string into an lxml tree object and split into segments

        :param html: The HTML string content of this field
        :type html: str
        """
        self.name = name
        if html_str := getattr(translation, name, ""):
            self.html = fromstring(html_str)
            # Translate all specified tags (and filter out empty segments)
            self.segments = [
                HTMLSegment(segment=segment)
                for segment in self.html.iter(*settings.SUMM_AI_HTML_TAGS)
            ]

    def __repr__(self):
        """
        The representation used for logging

        :return: The canonical string representation of the HTML field
        :rtype: str
        """
        return f"<HTMLField (segments: {self.segments})>"

    @property
    def translated_text(self):
        """
        Assemble the content of the HTML segments into a HTML string again

        :returns: The translated HTML
        :rtype: str
        """
        if self.html is not None:
            return tostring(
                self.html, encoding="unicode", method="html", pretty_print=True
            )
        return None

    @property
    def exception(self):
        """
        Check if any of the segments experienced an error

        :returns: The first exception of this HTML field
        :rtype: Exception
        """
        return next(
            (segment.exception for segment in self.segments if segment.exception), None
        )


# pylint: disable=too-many-instance-attributes
class TranslationHelper:
    """
    Custom helper class for interaction with SUMM.AI
    """

    #: The current request
    #:
    #: :type: ~django.http.HttpRequest
    request = None
    #: The :class:`~integreat_cms.cms.forms.custom_content_model_form.CustomContentModelForm`
    #: subclass of the current content type
    #:
    #: :type: ~django.forms.models.ModelFormMetaclass
    form_class = None
    #: The current object instance to be translated
    #:
    #: :type: ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel
    object_instance = None
    #: The German source translation of the object instance
    #:
    #: :type: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
    german_translation = None
    #: Wether or not the translation was successful
    #:
    #: :type: bool
    valid = True
    #: list: The text fields of this helper
    #:
    #: :type: list
    text_fields = []
    #: The HTML fields of this helper
    #:
    #: :type: list
    html_fields = []

    def __init__(self, request, form_class, object_instance):
        """
        Constructor initializes the class variables

        :param request: current request
        :type request: ~django.http.HttpRequest

        :param form_class: The :class:`~integreat_cms.cms.forms.custom_content_model_form.CustomContentModelForm`
                           subclass of the current content type
        :type form_class: :type: ~django.forms.models.ModelFormMetaclass

        :param object_instance: The current object instance
        :type object_instance: ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel
        """
        self.request = request
        self.form_class = form_class
        self.object_instance = object_instance
        self.german_translation = object_instance.get_translation(
            settings.SUMM_AI_GERMAN_LANGUAGE_SLUG
        )
        if not self.german_translation:
            messages.error(
                self.request,
                _('No German translation could be found for {} "{}".').format(
                    type(object_instance)._meta.verbose_name.title(),
                    object_instance.best_translation.title,
                ),
            )
            self.valid = False
            return
        self.text_fields = [
            TextField(name=text_field, translation=self.german_translation)
            for text_field in settings.SUMM_AI_TEXT_FIELDS
        ]
        self.html_fields = [
            HTMLField(name=html_field, translation=self.german_translation)
            for html_field in settings.SUMM_AI_HTML_FIELDS
        ]

    @property
    def fields(self):
        """
        Get all fields of this helper instance

        :returns: All fields which need to be translated
        :rtype: list [ ~integreat_cms.summ_ai_api.utils.TextField or ~integreat_cms.summ_ai_api.utils.HTMLField ]
        """
        return self.text_fields + self.html_fields

    def get_text_fields(self):
        """
        Get all text fields of this helper instance
        (all native :attr:`~integreat_cms.summ_ai_api.utils.TranslationHelper.text_fields`
        combined with all segments of all
        :attr:`~integreat_cms.summ_ai_api.utils.TranslationHelper.html_fields`)

        :returns: All text fields and segments which need to be translated
        :rtype: list [ ~integreat_cms.summ_ai_api.utils.TextField ]
        """
        if not self.valid:
            return []
        text_fields = list(
            itertools.chain(
                # Get all plain text fields
                [text_field for text_field in self.text_fields if text_field.text],
                # Get all segments of all HTML fields
                *[html_field.segments for html_field in self.html_fields],
            )
        )
        logger.debug(
            "Text fields for %r: %r",
            self,
            text_fields,
        )
        return text_fields

    def commit(self, easy_german):
        """
        Save the translated changes to the database

        :param easy_german: The language object of Easy German
        :type easy_german: ~integreat_cms.cms.models.languages.language.Language
        """
        if not self.valid:
            return
        # Check whether any of the fields returned an error
        if any(field.exception for field in self.fields):
            messages.error(
                self.request,
                __(
                    _(
                        '{} "{}" could not be automatically translated into Easy German.'
                    ).format(
                        type(self.object_instance)._meta.verbose_name.title(),
                        self.german_translation.title,
                    ),
                    _("Please try again later or contact an administrator"),
                ),
            )
            return
        # Initialize form to create new translation object
        content_translation_form = self.form_class(
            data={
                # Pass all inherited fields
                **{
                    field_name: getattr(self.german_translation, field_name, "")
                    for field_name in settings.SUMM_AI_INHERITED_FIELDS
                },
                # Pass all translated texts as data values
                **{field.name: field.translated_text for field in self.fields},
                # Always set automatic translations into pending review state
                "status": status.REVIEW,
            },
            instance=self.object_instance.get_translation(
                settings.SUMM_AI_EASY_GERMAN_LANGUAGE_SLUG
            ),
            additional_instance_attributes={
                "creator": self.request.user,
                "language": easy_german,
                self.german_translation.foreign_field(): self.object_instance,
            },
        )
        # Validate translation form
        if not content_translation_form.is_valid():
            logger.error(
                "Automatic translation into Easy German for %r could not be created because of %s",
                self.object_instance,
                content_translation_form.errors,
            )
            messages.error(
                self.request,
                _(
                    '{} "{}" could not be automatically translated into Easy German.'
                ).format(
                    type(self.object_instance)._meta.verbose_name.title(),
                    self.german_translation.title,
                ),
            )
            return
        # Save new translation
        content_translation_form.save()
        logger.debug(
            "Successfully translated %r into Easy German",
            content_translation_form.instance,
        )
        messages.success(
            self.request,
            _('{} "{}" has been successfully translated into Easy German.').format(
                type(self.object_instance)._meta.verbose_name.title(),
                self.german_translation.title,
            ),
        )

    def __repr__(self):
        """
        The representation used for logging

        :return: The canonical string representation of the translation helper
        :rtype: str
        """
        return f"<TranslationHelper (translation: {self.german_translation!r})>"



import logging

from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import render, redirect
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView

from ...constants import status
from ...decorators import permission_required
from ...models import Language, ImprintPage

logger = logging.getLogger(__name__)


@method_decorator(permission_required("cms.view_imprintpage"), name="dispatch")
@method_decorator(permission_required("cms.change_imprintpage"), name="post")
class ImprintRevisionView(TemplateView):
    """
    View for browsing the imprint revisions and restoring old imprint revisions
    """

    template_name = "imprint/imprint_revisions.html"
    #: The context dict passed to the template (see :class:`~django.views.generic.base.ContextMixin`)
    extra_context = {"current_menu_item": "imprint"}

    def get(self, request, *args, **kwargs):
        r"""
        Render imprint page revision slider

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.http.Http404: If no imprint exists for the region

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region
        imprint = region.imprint
        if not imprint:
            raise Http404("No imprint found for this region")

        language = region.get_language_or_404(
            kwargs.get("language_slug"), only_active=True
        )

        imprint_translations = imprint.translations.filter(language=language)

        if not imprint_translations.exists():
            return redirect(
                "edit_imprint",
                **{
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )

        selected_revision = imprint_translations.filter(
            version=kwargs.get("selected_revision", imprint_translations.count())
        )

        if not selected_revision.exists():
            messages.error(request, _("This revision does not exist."))
            return redirect(
                "imprint_revisions",
                **{
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )

        # Show warning if user has no permission to manage the imprint
        if not request.user.has_perm("cms.change_imprintpage"):
            messages.warning(
                request,
                _("You don't have the permission restore revisions of the imprint."),
            )

        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "imprint": imprint,
                "imprint_translations": imprint_translations,
                "api_revision": imprint_translations.filter(
                    status=status.PUBLIC
                ).first(),
                "selected_revision": selected_revision.first(),
                "language": language,
            },
        )

    # pylint: disable=unused-argument
    def post(self, request, *args, **kwargs):
        r"""
        Restore a previous revision of an imprint page translation

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.http.Http404: If no imprint exists for the region

        :raises ~django.core.exceptions.PermissionDenied: If user tries to restore a page with an invalid status

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region
        try:
            imprint = region.imprint
        except ImprintPage.DoesNotExist as e:
            raise Http404 from e

        language = Language.objects.get(slug=kwargs.get("language_slug"))

        revision = imprint.translations.filter(
            language=language, version=request.POST.get("revision")
        ).first()

        if not revision:
            messages.error(request, _("This revision does not exist."))
            return redirect(
                "imprint_revisions",
                **{
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )

        current_revision = imprint.get_translation(language.slug)

        if (
            revision.title == current_revision.title
            and revision.content == current_revision.content
        ):
            messages.error(
                request,
                _(
                    "This revision is identical to the current version of this translation."
                ),
            )
            return redirect(
                "imprint_revisions",
                **{
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )

        revision.pk = None
        revision.version = current_revision.version + 1
        # Reset author to current user
        revision.creator = request.user

        desired_status = request.POST.get("status")

        if desired_status not in dict(status.CHOICES):
            raise PermissionDenied(
                f"{request.user!r} tried to restore {revision!r} of {imprint!r} with invalid status {desired_status!r}"
            )

        revision.status = desired_status
        revision.save()

        messages.success(request, _("The revision was successfully restored"))

        imprint_translations = imprint.translations.filter(language=language)

        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "imprint": imprint,
                "imprint_translations": imprint_translations,
                "api_revision": imprint_translations.filter(
                    status=status.PUBLIC
                ).first(),
                "selected_revision": revision,
                "language": language,
            },
        )



import logging
import json
from webauthn import verify_registration_response, base64url_to_bytes
from webauthn.helpers.structs import RegistrationCredential

from django.conf import settings
from django.contrib import messages
from django.http import JsonResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic import CreateView

from ....models import UserMfaKey
from ....decorators import modify_mfa_authenticated

logger = logging.getLogger(__name__)


@method_decorator(modify_mfa_authenticated, name="dispatch")
class RegisterUserMfaKeyView(CreateView):
    """
    View to render a form for creating :class:`~integreat_cms.cms.models.users.user_mfa_key.UserMfaKey` objects
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "settings/mfa/add_key.html"
    #: The model of this :class:`~django.views.generic.edit.CreateView`
    model = UserMfaKey
    #: The fields of the model which should be handled by this create view
    fields = [
        "name",
    ]

    def post(self, request, *args, **kwargs):
        r"""
        Verify a registration challenge and register a 2-FA key.
        Called asynchronously by JavaScript.

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The JSON response
        :rtype: ~django.http.JsonResponse
        """
        json_data = json.loads(request.body)

        webauthn_registration_response = verify_registration_response(
            credential=RegistrationCredential.parse_raw(request.body),
            expected_rp_id=settings.HOSTNAME,
            expected_origin=settings.BASE_URL,
            expected_challenge=base64url_to_bytes(
                request.session["mfa_registration_challenge"]
            ),
        )

        existing_key = request.user.mfa_keys.filter(name=json_data["name"])
        if existing_key.exists():
            return JsonResponse(
                {"success": False, "error": _("This key name has already been used")}
            )

        existing_key = request.user.mfa_keys.filter(
            key_id=webauthn_registration_response.credential_id
        )
        if existing_key.exists():
            return JsonResponse(
                {"success": False, "error": _("You already registered this key")}
            )

        new_key = UserMfaKey(
            user=request.user,
            name=json_data["name"],
            key_id=webauthn_registration_response.credential_id,
            public_key=webauthn_registration_response.credential_public_key,
            sign_count=webauthn_registration_response.sign_count,
        )
        new_key.save()
        messages.success(
            request,
            _(
                'The 2-factor authentication key "{}" was successfully registered.'
            ).format(new_key.name),
        )
        # Determine success url based on current region
        kwargs = {"region_slug": request.region.slug} if request.region else {}
        success_url = reverse("user_settings", kwargs=kwargs)
        return JsonResponse({"success": True, "successUrl": success_url})



"""
This file contains custom authentication backends, see :ref:`ref/contrib/auth:authentication backends` and
:ref:`django:authentication-backends`.
"""
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import BaseBackend


UserModel = get_user_model()


class EmailAuthenticationBackend(BaseBackend):
    """
    This authentication backend allows users to login with their email address instead of their username
    """

    def authenticate(self, request, **kwargs):
        r"""
        Try to authenticate a user with the given email and password

        :param request: The current request
        :type request: ~django.http.HttpRequest

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: Either the authenticated user or ``None`` is the credentials were not valid
        :rtype: ~integreat_cms.cms.models.users.user.User
        """
        email = kwargs.get(UserModel.USERNAME_FIELD)
        if not email:
            return None
        password = kwargs.get("password")
        try:
            user = UserModel.objects.get(**{UserModel.EMAIL_FIELD: email.lower()})
        except UserModel.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a nonexistent user (#20760).
            UserModel().set_password(password)
        else:
            if user.check_password(password) and user.is_active:
                return user
        return None

    def get_user(self, user_id):
        """
        Get the user by its primary key

        :param user_id: The id of the user
        :type user_id: int

        :return: Either the user or ``None`` if no user with this id exists
        :rtype: ~integreat_cms.cms.models.users.user.User
        """
        try:
            return UserModel.objects.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None



"""
Context processors pass additional variables to templates (see :ref:`context-processors`).
"""
from django.conf import settings

from .. import __version__
from ..cms.constants import status


# pylint: disable=unused-variable
def version_processor(request):
    """
    This context processor injects the current package version into the template context.

    :param request: The current http request
    :type request: ~django.http.HttpRequest

    :return: The currently installed version of this package
    :rtype: dict
    """
    return {"version": __version__}


# pylint: disable=unused-variable
def settings_processor(request):
    """
    This context processor injects a few settings into the template context.

    :param request: The current http request
    :type request: ~django.http.HttpRequest

    :return: A few of our settings
    :rtype: dict
    """
    return {
        "FCM_ENABLED": settings.FCM_ENABLED,
        "BRANDING": settings.BRANDING,
        "WEBAPP_URL": settings.WEBAPP_URL,
        "DEEPL_ENABLED": settings.DEEPL_ENABLED,
    }


# pylint: disable=unused-variable
def constants_processor(request):
    """
    This context processor injects some of our constants into the template context.

    :param request: The current http request
    :type request: ~django.http.HttpRequest

    :return: A few of our constants
    :rtype: dict
    """
    return {
        "DRAFT": status.DRAFT,
        "REVIEW": status.REVIEW,
        "AUTO_SAVE": status.AUTO_SAVE,
        "PUBLIC": status.PUBLIC,
    }



"""
This module contains mixins for our views
"""
from django.conf import settings
from django.contrib.auth.mixins import UserPassesTestMixin
from django.views.generic.base import ContextMixin, TemplateResponseMixin
from django.utils.translation import ugettext as _
from django.urls import reverse


class RegionPermissionRequiredMixing(UserPassesTestMixin):
    """
    A mixin that can be used for class-based views that require that the user is has access to the current region of this request
    """

    def test_func(self):
        """
        The test the user has to pass

        :return: Whether the user has passed the test
        :rtype: bool
        """
        # Superusers and staff have permissions for all regions
        return (
            self.request.user.is_superuser
            or self.request.user.is_staff
            or self.request.region in self.request.user.regions.all()
        )


class ModelTemplateResponseMixin(TemplateResponseMixin):
    """
    A mixin that can be used to render a template of a model (e.g. list or form)
    """

    @property
    def template_name(self):
        """
        Return the template name to be used for the request.

        :return: The template to be rendered
        :rtype: str
        """
        model_name = self.model._meta.model_name
        return f"{model_name}s/{model_name}{self.template_name_suffix}.html"


# pylint: disable=too-few-public-methods
class ModelConfirmationContextMixin(ContextMixin):
    """
    A mixin that can be used to inject confirmation text into a template of a model (e.g. list or form)
    """

    def get_context_data(self, **kwargs):
        r"""
        Returns a dictionary representing the template context
        (see :meth:`~django.views.generic.base.ContextMixin.get_context_data`).

        :param \**kwargs: The given keyword arguments
        :type \**kwargs: dict

        :return: The template context
        :rtype: dict
        """
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "delete_dialog_title": _(
                    "Please confirm that you really want to delete this {}"
                ).format(self.model._meta.verbose_name),
                "delete_dialog_text": _("This cannot be reversed."),
            }
        )
        return context


class ContentEditLockMixin(ContextMixin):
    """
    A mixin that provides some variables required for the content edit lock
    """

    #: The reverse name of the url of the associated list view
    back_url_name = None

    def get_context_data(self, **kwargs):
        r"""
        Returns a dictionary representing the template context
        (see :meth:`~django.views.generic.base.ContextMixin.get_context_data`).

        :param \**kwargs: The given keyword arguments
        :type \**kwargs: dict

        :return: The template context
        :rtype: dict
        """
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "back_url": reverse(
                    self.back_url_name,
                    kwargs={
                        "region_slug": kwargs["region_slug"],
                        "language_slug": kwargs["language_slug"],
                    },
                )
            }
        )
        return context


# pylint: disable=too-few-public-methods
class SummAiContextMixin(ContextMixin):
    """
    This mixin provides extra context for SUMM.AI bulk action views
    """

    def get_context_data(self, **kwargs):
        r"""
        Returns a dictionary representing the template context
        (see :meth:`~django.views.generic.base.ContextMixin.get_context_data`).

        :param \**kwargs: The given keyword arguments
        :type \**kwargs: dict

        :return: The template context
        :rtype: dict
        """
        context = super().get_context_data(**kwargs)
        context["SUMM_AI_ENABLED"] = (
            settings.SUMM_AI_ENABLED
            and self.request.region.summ_ai_enabled
            and kwargs.get("language_slug")
            in [
                settings.SUMM_AI_GERMAN_LANGUAGE_SLUG,
                settings.SUMM_AI_EASY_GERMAN_LANGUAGE_SLUG,
            ]
        )
        return context



import pytest

from ..utils import check_view_status_code
from ..view_config import PARAMETRIZED_VIEWS


# pylint: disable=missing-function-docstring
@pytest.mark.django_db
@pytest.mark.parametrize("view_name,kwargs,post_data,roles", PARAMETRIZED_VIEWS[4::16])
def test_view_status_code_4(login_role_user, view_name, kwargs, post_data, roles):
    check_view_status_code(login_role_user, view_name, kwargs, post_data, roles)



import logging

from django.conf import settings
from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseNotFound,
    HttpResponseServerError,
)
from django.template.loader import render_to_string
from django.utils.translation import ugettext as _

logger = logging.getLogger(__name__)


def render_error_template(context):
    """
    Render the HTTP error template

    :param context: The context data for the error template
    :type context: dict

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    context.update(
        {
            "COMPANY": settings.COMPANY,
            "COMPANY_URL": settings.COMPANY_URL,
            "BRANDING": settings.BRANDING,
        }
    )
    return render_to_string("error_handler/http_error.html", context)


def handler400(request, exception):
    """
    Render a HTTP 400 Error code

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param exception: Exception (unused)
    :type exception: BaseException

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    context = {
        "request": request,
        "code": 400,
        "title": _("Bad request"),
        "message": _("There was an error in your request."),
    }
    logger.debug(exception)
    return HttpResponseBadRequest(render_error_template(context))


def handler403(request, exception):
    """
    Render a HTTP 403 Error code

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param exception: Exception (unused)
    :type exception: BaseException

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    context = {
        "request": request,
        "code": 403,
        "title": _("Forbidden"),
        "message": _("You don't have the permission to access this page."),
    }
    logger.debug(exception)
    return HttpResponseForbidden(render_error_template(context))


def handler404(request, exception):
    """
    Render a HTTP 404 Error code

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param exception: Exception (unused)
    :type exception: BaseException

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    context = {
        "request": request,
        "code": 404,
        "title": _("Page not found"),
        "message": _("The page you requested could not be found."),
    }
    logger.debug(exception)
    return HttpResponseNotFound(render_error_template(context))


def handler500(request):
    """
    Render a HTTP 500 Error code

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    context = {
        "request": request,
        "code": 500,
        "title": _("Internal Server Error"),
        "message": _("An unexpected error has occurred."),
    }
    return HttpResponseServerError(render_error_template(context))


def csrf_failure(request, reason):
    """
    Render a CSRF failure notice

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param reason: Description of reason for CSRF failure
    :type reason: str

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    context = {
        "request": request,
        "code": 403,
        "title": _("CSRF Error"),
        "message": _("Please try to reload the page."),
    }
    logger.debug(reason)
    return HttpResponseForbidden(render_error_template(context))



import logging

from django.conf import settings
from django.contrib import messages
from django.shortcuts import render, redirect
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView

from ...decorators import permission_required
from ...forms import PageFilterForm
from ..mixins import SummAiContextMixin
from .page_context_mixin import PageContextMixin

logger = logging.getLogger(__name__)


@method_decorator(permission_required("cms.view_page"), name="dispatch")
class PageTreeView(TemplateView, PageContextMixin, SummAiContextMixin):
    """
    View for showing the page tree
    """

    #: Template for list of non-archived pages
    template = "pages/page_tree.html"
    #: Template for list of archived pages
    template_archived = "pages/page_tree_archived.html"
    #: Whether or not to show archived pages
    archived = False

    @property
    def template_name(self):
        """
        Select correct HTML template, depending on :attr:`~integreat_cms.cms.views.pages.page_tree_view.PageTreeView.archived` flag
        (see :class:`~django.views.generic.base.TemplateResponseMixin`)

        :return: Path to HTML template
        :rtype: str
        """

        return self.template_archived if self.archived else self.template

    def get(self, request, *args, **kwargs):
        r"""
        Render page tree

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        # current region
        region = request.region

        # current language
        language_slug = kwargs.get("language_slug")
        if language_slug:
            language = region.get_language_or_404(language_slug, only_active=True)
        elif region.default_language:
            return redirect(
                "pages",
                **{
                    "region_slug": region.slug,
                    "language_slug": region.default_language.slug,
                }
            )
        else:
            messages.error(
                request,
                _("Please create at least one language node before creating pages."),
            )
            return redirect(
                "language_tree",
                **{
                    "region_slug": region.slug,
                }
            )

        if not request.user.has_perm("cms.change_page"):
            messages.warning(
                request, _("You don't have the permission to edit or create pages.")
            )

        # Initialize page filter form
        filter_form = PageFilterForm(data=request.GET)

        if filter_form.is_enabled or self.archived:
            # If filters are applied or only archived pages are requested, fetch all elements from the database
            page_queryset = region.pages.all()
        else:
            # Else, only fetch the root pages and load the subpages dynamically via ajax
            page_queryset = region.pages.filter(lft=1)

        # Cache tree structure to reduce database queries
        pages = page_queryset.prefetch_major_public_translations().cache_tree(
            archived=self.archived
        )

        # Filter pages according to given filters, if any
        pages = filter_form.apply(pages, language_slug)

        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "pages": pages,
                "language": language,
                "languages": region.active_languages,
                "filter_form": filter_form,
                "XLIFF_EXPORT_VERSION": settings.XLIFF_EXPORT_VERSION,
            },
        )



import pytest

from django.test.client import Client
from django.urls import reverse

from .view_config import PARAMETRIZED_PUBLIC_VIEWS


# pylint: disable=unused-argument
@pytest.mark.django_db
@pytest.mark.parametrize("view_name,post_data", PARAMETRIZED_PUBLIC_VIEWS)
def test_public_view_status_code(load_test_data, view_name, post_data):
    """
    This test checks whether the given view return the correct status code for anonymous users

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: NoneType

    :param view_name: The identifier of the view
    :type view_name: str

    :param post_data: The post data for this view
    :type post_data: dict
    """
    client = Client()
    url = reverse(view_name)
    if post_data:
        response = client.post(url, data=post_data)
    else:
        response = client.get(url)
    print(response.headers)
    if post_data:
        # Post-views should redirect after a successful operation
        assert response.status_code == 302
    else:
        # Get-views should return 200
        assert response.status_code == 200



import logging

from django.contrib import messages
from django.contrib.auth import views as auth_views
from django.utils.translation import ugettext as _
from django.http import HttpResponseRedirect
from django.urls import reverse_lazy
from django.shortcuts import redirect

logger = logging.getLogger(__name__)


class PasswordResetConfirmView(auth_views.PasswordResetConfirmView):
    """
    View to confirm that the password should be reset
    """

    #: The template which should be rendered
    template_name = "authentication/password_reset_confirm.html"
    #: If the password was successfully reset, redirect to the login
    success_url = reverse_lazy("public:login")

    def dispatch(self, *args, **kwargs):
        r"""
        The view part of the view. Handles all HTTP methods equally.

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response or a redirection to the password reset page
        :rtype: ~django.template.response.TemplateResponse or ~django.http.HttpResponseRedirect
        """
        if self.request.user.is_authenticated:
            messages.success(
                self.request,
                _("You are already logged in."),
            )
            return redirect("public:region_selection")
        response = super().dispatch(*args, **kwargs)
        if isinstance(response, HttpResponseRedirect) or self.validlink:
            # If the link is valid, render the password reset confirm form (redirect means valid because the first step
            # is to store the token in a session variable and redirect to the generic [...]-set-password/ url)
            return response
        # If not, redirect to the password reset form
        messages.error(
            self.request,
            " ".join(
                [
                    _("This password reset link is invalid."),
                    _("It may have already been used."),
                    _("Please request a new link to reset your password."),
                ]
            ),
        )
        logger.debug("An invalid password reset link was used.")
        return redirect("public:password_reset")

    def form_valid(self, form):
        """
        If the form is valid, show a success message.

        :param form: The supplied form
        :type form: ~django.contrib.auth.forms.SetPasswordForm

        :return: form validation
        :rtype: ~django.http.HttpResponse

        """
        messages.success(
            self.request,
            " ".join(
                [
                    _("Your password has been successfully changed."),
                    _("You can now log in with your new password."),
                ]
            ),
        )
        logger.info("The password for %r was changed.", form.user)
        return super().form_valid(form)



from django.contrib.auth import get_user_model
from django.db.models import Q


def search_users(region, query):
    """
    Searches for all users that match the given `query`.
    If region is None, all users are searched.
    :param region: The current region
    :type region: ~integreat_cms.cms.models.regions.region.Region
    :param query: The query string used for filtering the regions
    :type query: str
    :return: A query for all matching objects
    :rtype: ~django.db.models.QuerySet
    """

    filter_query = (
        Q(username__icontains=query)
        | Q(first_name__icontains=query)
        | Q(last_name__icontains=query)
    )

    if region:
        objects = region.region_users
    else:
        objects = get_user_model().objects

    return objects.filter(filter_query)



import logging

from ...models import PageTranslation
from ..custom_content_model_form import CustomContentModelForm


logger = logging.getLogger(__name__)


class PageTranslationForm(CustomContentModelForm):
    """
    Form for creating and modifying page translation objects
    """

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = PageTranslation
        #: The fields of the model which should be handled by this form
        fields = ["title", "slug", "status", "content", "minor_edit"]



import pytest

from django.test.client import Client
from django.urls import resolve, reverse

from integreat_cms.cms.models import Page


# pylint: disable=unused-argument
@pytest.mark.django_db
def test_api_push_page_content(load_test_data):
    """
    This test class checks all endpoints defined in :attr:`~tests.api.api_config.API_ENDPOINTS`.
    It verifies that the content delivered by the endpoint is equivalent with the data
    provided in the corresponding json file.

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: tuple
    """
    client = Client()
    valid_token = "valid-token"
    invalid_token = "invalid-token"
    new_content = "<p>new content</p>"
    view_name = "api:push_page_translation_content"
    endpoint = reverse(
        view_name, kwargs={"region_slug": "augsburg", "language_slug": "de"}
    )
    # Check whether the endpoints resolve correctly
    match = resolve("/api/augsburg/de/pushpage/")
    wp_match = resolve("/augsburg/de/wp-json/extensions/v3/pushpage/")
    assert match.view_name == wp_match.view_name == view_name
    # Test invalid submission
    response = client.post(
        endpoint,
        {"content": new_content, "token": invalid_token},
        format="json",
        content_type="application/json",
    )
    print(response.headers)
    assert response.status_code == 403
    assert response.json() == {"status": "denied"}
    assert not Page.objects.filter(api_token=invalid_token).exists()
    # Test valid submission
    page = Page.objects.get(api_token=valid_token)
    assert page.get_translation("de").content != new_content
    response = client.post(
        endpoint,
        {"content": new_content, "token": valid_token},
        format="json",
        content_type="application/json",
    )
    print(response.headers)
    assert response.status_code == 200
    assert response.json() == {"status": "success"}
    page = Page.objects.get(api_token=valid_token)
    assert page.get_translation("de").content == new_content



"""
This module contains action methods for feedback items (archive, restore, ...)
"""
import logging

from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from cacheops import invalidate_obj

from ...decorators import permission_required
from ...models import Feedback

logger = logging.getLogger(__name__)


@require_POST
@permission_required("cms.change_feedback")
def mark_region_feedback_as_read(request, region_slug):
    """
    Set read flag for a list of feedback items

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param region_slug: The slug of the current region
    :type region_slug: str

    :return: A redirection to the region feedback list
    :rtype: ~django.http.HttpResponseRedirect
    """

    region = request.region

    selected_ids = request.POST.getlist("selected_ids[]")
    selected_feedback = Feedback.objects.filter(
        id__in=selected_ids, region=region, is_technical=False
    )
    for feedback in selected_feedback:
        invalidate_obj(feedback)
        if hasattr(feedback, "feedback_ptr"):
            invalidate_obj(feedback.feedback_ptr)
    selected_feedback.update(read_by=request.user)

    logger.debug(
        "Feedback objects %r marked as read by %r",
        selected_ids,
        request.user,
    )
    messages.success(request, _("Feedback was successfully marked as read"))

    return redirect("region_feedback", region_slug=region_slug)


@require_POST
@permission_required("cms.change_feedback")
def mark_region_feedback_as_unread(request, region_slug):
    """
    Unset read flag for a list of feedback items

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param region_slug: The slug of the current region
    :type region_slug: str

    :return: A redirection to the region feedback list
    :rtype: ~django.http.HttpResponseRedirect
    """

    region = request.region

    selected_ids = request.POST.getlist("selected_ids[]")
    selected_feedback = Feedback.objects.filter(
        id__in=selected_ids, region=region, is_technical=False
    )
    for feedback in selected_feedback:
        invalidate_obj(feedback)
        if hasattr(feedback, "feedback_ptr"):
            invalidate_obj(feedback.feedback_ptr)
    selected_feedback.update(read_by=None)

    logger.debug(
        "Feedback objects %r marked as unread by %r",
        selected_ids,
        request.user,
    )
    messages.success(request, _("Feedback was successfully marked as unread"))

    return redirect("region_feedback", region_slug=region_slug)


@require_POST
@permission_required("cms.delete_feedback")
def delete_region_feedback(request, region_slug):
    """
    Delete a list of feedback items

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param region_slug: The slug of the current region
    :type region_slug: str

    :return: A redirection to the region feedback list
    :rtype: ~django.http.HttpResponseRedirect
    """

    region = request.region

    selected_ids = request.POST.getlist("selected_ids[]")
    Feedback.objects.filter(
        id__in=selected_ids, region=region, is_technical=False
    ).delete()

    logger.info("Feedback objects %r deleted by %r", selected_ids, request.user)
    messages.success(request, _("Feedback was successfully deleted"))

    return redirect("region_feedback", region_slug=region_slug)



"""
This is a collection of tags and filters for strings.
"""
import json

from django import template
from django.utils.translation import ugettext as _

register = template.Library()


@register.filter(name="words")
def words(text):
    """
    Split the given text into a list of words, see :meth:`python:str.split`.

    :param text: The input string
    :type text: str

    :return: The list of words in the text
    :rtype: list
    """
    return text.split()


@register.filter(name="linkcheck_status_filter")
def linkcheck_status_filter(status_message):
    """
    Due to a long status entry for a single kind of faulty link,
    this filter reduced the output when display in list view

    :param status_message: error description
    :type status_message: str
    :return: a concise message
    :rtype: str
    """
    if not status_message:
        return _("Unknown")
    if status_message.startswith("Other Error:"):
        return _("Error")
    # Sometimes 404 errors are malformed
    if status_message in ["404 ", "404 404"]:
        return "404 Not Found"
    return _(status_message)


@register.filter(name="to_json")
def to_json(obj):
    """
    Converts the given object to a json string

    :param obj: The input object
    :type obj: object

    :return: object as json string
    :rtype: str
    """
    return json.dumps(obj)



"""
This modules contains the config for the API tests
"""

#: The API endpoints
API_ENDPOINTS = [
    (
        "/api/regions/",
        "/wp-json/extensions/v3/sites/",
        "tests/api/expected-outputs/regions.json",
        200,
    ),
    (
        "/api/regions/live/",
        "/wp-json/extensions/v3/sites/live/",
        "tests/api/expected-outputs/regions_live.json",
        200,
    ),
    (
        "/api/augsburg/de/pages/",
        "/augsburg/de/wp-json/extensions/v3/pages/",
        "tests/api/expected-outputs/augsburg_de_pages.json",
        200,
    ),
    (
        "/api/augsburg/ar/pages/",
        "/augsburg/ar/wp-json/extensions/v3/pages/",
        "tests/api/expected-outputs/augsburg_ar_pages.json",
        200,
    ),
    (
        "/api/augsburg/non-existing/pages/",
        "/augsburg/non-existing/wp-json/extensions/v3/pages/",
        "tests/api/expected-outputs/augsburg_non-existing_pages.json",
        404,
    ),
    (
        "/api/augsburg/de/locations/",
        "/augsburg/de/wp-json/extensions/v3/locations/",
        "tests/api/expected-outputs/augsburg_de_locations.json",
        200,
    ),
    (
        "/api/augsburg/en/locations/",
        "/augsburg/en/wp-json/extensions/v3/locations/",
        "tests/api/expected-outputs/augsburg_en_locations.json",
        200,
    ),
    (
        "/api/augsburg/ar/locations/",
        "/augsburg/ar/wp-json/extensions/v3/locations/",
        "tests/api/expected-outputs/augsburg_ar_locations.json",
        200,
    ),
    (
        "/api/augsburg/non-existing/locations/",
        "/augsburg/non-existing/wp-json/extensions/v3/locations/",
        "tests/api/expected-outputs/augsburg_non-existing_locations.json",
        404,
    ),
    (
        "/api/augsburg/de/children/",
        "/augsburg/de/wp-json/extensions/v3/children/",
        "tests/api/expected-outputs/augsburg_de_children.json",
        200,
    ),
    (
        "/api/augsburg/de/children/?depth=3&url=/augsburg/de/behörden-und-beratung/behörden/",
        "/augsburg/de/wp-json/extensions/v3/children/?depth=3&url=/augsburg/de/behörden-und-beratung/behörden/",
        "tests/api/expected-outputs/augsburg_de_children_archived_descendants.json",
        200,
    ),
    (
        "/augsburg/de/wp-json/extensions/v3/page/?url=/augsburg/de/behörden-und-beratung/behörden/archiviertes-amt/",
        "/augsburg/de/wp-json/extensions/v3/page/?url=/augsburg/de/behörden-und-beratung/behörden/archiviertes-amt/",
        "tests/api/expected-outputs/augsburg_de_page_archived.json",
        404,
    ),
    (
        "/api/augsburg/de/children/?depth=2",
        "/augsburg/de/wp-json/extensions/v3/children/?depth=2",
        "tests/api/expected-outputs/api_augsburg_de_children_depth_2.json",
        200,
    ),
    (
        "/api/augsburg/de/events/",
        "/augsburg/de/wp-json/extensions/v3/events/",
        "tests/api/expected-outputs/augsburg_de_events.json",
        200,
    ),
    (
        "/api/augsburg/en/events/",
        "/augsburg/en/wp-json/extensions/v3/events/",
        "tests/api/expected-outputs/augsburg_en_events.json",
        200,
    ),
    (
        "/api/augsburg/ar/events/",
        "/augsburg/ar/wp-json/extensions/v3/events/",
        "tests/api/expected-outputs/augsburg_ar_events.json",
        200,
    ),
    (
        "/api/augsburg/non-existing/events/",
        "/augsburg/non-existing/wp-json/extensions/v3/events/",
        "tests/api/expected-outputs/augsburg_non-existing_events.json",
        404,
    ),
    (
        "/api/augsburg/de/page/?id=1",
        "/augsburg/de/wp-json/extensions/v3/page/?id=1",
        "tests/api/expected-outputs/augsburg_de_page_1.json",
        200,
    ),
    (
        "/api/augsburg/de/imprint/",
        "/augsburg/de/wp-json/extensions/v3/disclaimer/",
        "tests/api/expected-outputs/augsburg_de_imprint.json",
        200,
    ),
    (
        "/api/augsburg/en/imprint/",
        "/augsburg/en/wp-json/extensions/v3/disclaimer/",
        "tests/api/expected-outputs/augsburg_en_imprint.json",
        200,
    ),
    (
        "/api/augsburg/non-existing/imprint/",
        "/augsburg/non-existing/wp-json/extensions/v3/disclaimer/",
        "tests/api/expected-outputs/augsburg_non-existing_imprint.json",
        404,
    ),
    (
        "/api/nurnberg/de/events/",
        "/nurnberg/de/wp-json/extensions/v3/events/",
        "tests/api/expected-outputs/nurnberg_de_events.json",
        200,
    ),
    (
        "/api/nurnberg/en/events/",
        "/nurnberg/en/wp-json/extensions/v3/events/",
        "tests/api/expected-outputs/nurnberg_en_events.json",
        200,
    ),
    (
        "/api/nurnberg/ar/events/",
        "/nurnberg/ar/wp-json/extensions/v3/events/",
        "tests/api/expected-outputs/nurnberg_ar_events.json",
        200,
    ),
    (
        "/api/nurnberg/de/locations/",
        "/nurnberg/de/wp-json/extensions/v3/locations/",
        "tests/api/expected-outputs/nurnberg_de_locations.json",
        200,
    ),
    (
        "/api/nurnberg/en/locations/",
        "/nurnberg/en/wp-json/extensions/v3/locations/",
        "tests/api/expected-outputs/nurnberg_en_locations.json",
        200,
    ),
    (
        "/api/nurnberg/ar/locations/",
        "/nurnberg/ar/wp-json/extensions/v3/locations/",
        "tests/api/expected-outputs/nurnberg_ar_locations.json",
        200,
    ),
    (
        "/api/nurnberg/de/imprint/",
        "/nurnberg/de/wp-json/extensions/v3/disclaimer/",
        "tests/api/expected-outputs/nurnberg_de_imprint.json",
        200,
    ),
    (
        "/api/nurnberg/en/imprint/",
        "/nurnberg/en/wp-json/extensions/v3/disclaimer/",
        "tests/api/expected-outputs/nurnberg_en_imprint.json",
        200,
    ),
    (
        "/api/nurnberg/de/fcm/",
        "/nurnberg/de/wp-json/extensions/v3/fcm/",
        "tests/api/expected-outputs/nurnberg_de_fcm.json",
        200,
    ),
    (
        "/api/nurnberg/en/fcm/",
        "/nurnberg/en/wp-json/extensions/v3/fcm/",
        "tests/api/expected-outputs/nurnberg_en_fcm.json",
        200,
    ),
    (
        "/api/nurnberg/ar/fcm/",
        "/nurnberg/ar/wp-json/extensions/v3/fcm/",
        "tests/api/expected-outputs/nurnberg_ar_fcm.json",
        200,
    ),
    (
        "/api/augsburg/de/extras/",
        "/augsburg/de/wp-json/extensions/v3/extras/",
        "tests/api/expected-outputs/augsburg-offers.json",
        200,
    ),
]

#: This list contains the config for all API feedback views
API_FEEDBACK_VIEWS = [
    (
        "api:legacy_feedback_endpoint",
        {
            "permalink": "/augsburg/de/willkommen",
            "comment": "Cool page!",
            "rating": "up",
            "category": "Inhalte",
        },
    ),
    (
        "api:legacy_feedback_endpoint",
        {
            "permalink": "/augsburg/de/willkommen",
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:legacy_feedback_endpoint",
        {
            "permalink": "/augsburg/de/events/test-veranstaltung",
            "comment": "Cool event!",
            "rating": "up",
            "category": "Inhalte",
        },
    ),
    (
        "api:legacy_feedback_endpoint",
        {
            "permalink": "/augsburg/de/events/test-veranstaltung",
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:region_feedback",
        {"comment": "Cool region!", "rating": "up", "category": "Inhalte"},
    ),
    (
        "api:region_feedback",
        {
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:page_feedback",
        {
            "slug": "willkommen",
            "comment": "Cool page!",
            "rating": "up",
            "category": "Inhalte",
        },
    ),
    (
        "api:page_feedback",
        {
            "slug": "willkommen",
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:poi_feedback",
        {
            "slug": "test-ort",
            "comment": "Cool POI!",
            "rating": "up",
            "category": "Inhalte",
        },
    ),
    (
        "api:poi_feedback",
        {
            "slug": "test-ort",
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:event_feedback",
        {
            "slug": "test-veranstaltung",
            "comment": "Cool event!",
            "rating": "up",
            "category": "Inhalte",
        },
    ),
    (
        "api:event_feedback",
        {
            "slug": "test-veranstaltung",
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:event_list_feedback",
        {"comment": "Cool events!", "rating": "up", "category": "Inhalte"},
    ),
    (
        "api:event_list_feedback",
        {
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:imprint_page_feedbacks",
        {"comment": "Cool imprint!", "rating": "up", "category": "Inhalte"},
    ),
    (
        "api:imprint_page_feedbacks",
        {
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:map_feedback",
        {"comment": "Cool map!", "rating": "up", "category": "Inhalte"},
    ),
    (
        "api:map_feedback",
        {
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:search_result_feedback",
        {
            "query": "search query",
            "comment": "Cool search results!",
            "rating": "up",
            "category": "Inhalte",
        },
    ),
    (
        "api:search_result_feedback",
        {
            "query": "search query",
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:offer_list_feedback",
        {"comment": "Cool offers!", "rating": "up", "category": "Inhalte"},
    ),
    (
        "api:offer_list_feedback",
        {
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
    (
        "api:offer_feedback",
        {
            "slug": "sprungbrett",
            "comment": "Cool offer!",
            "rating": "up",
            "category": "Inhalte",
        },
    ),
    (
        "api:offer_feedback",
        {
            "slug": "sprungbrett",
            "comment": "Strange bug!",
            "rating": "down",
            "category": "Technisches Feedback",
        },
    ),
]



"""
This module includes functions related to the offers API endpoint.
"""
from django.http import JsonResponse

from ...cms.constants import postal_code
from ..decorators import json_response


def get_url(offer, region):
    """
    The offer should inherit the slug property from its template. This is the url to an API endpoint in most cases.
    Some offers depend on the location which is realized by adding the postal code of the current region to the
    request. If the offer template indicates that the postal code should be used as ``GET``-parameter, the class
    attribute ``use_postal_code`` has to be set to ``postal_code.GET`` (see :mod:`~integreat_cms.cms.constants.postal_code`) and
    the url has to end with the name of the required parameter-name, e.g. ``https://example.com/api?location=``.

    :param offer: one offer (formerly extra)
    :type offer: ~integreat_cms.cms.models.offers.offer_template.OfferTemplate

    :param region: current region object
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :return: The url of an offer
    :rtype: str
    """
    if offer.use_postal_code == postal_code.GET:
        return offer.url + region.postal_code
    return offer.url


def get_post_data(offer, region):
    """
    In case the url expects additional post data, it is stored inside the ``post_data``-dict. Some offers depend on
    the location which is realized by adding the postal code of the current region to the request. If the offer
    template indicates that the postal code should be used as ``GET``-parameter, the class attribute
    ``use_postal_code`` has to be set to ``postal_code.POST`` (see :mod:`~integreat_cms.cms.constants.postal_code`) and then the
    key ``search-plz`` is automatically added to the post data. In case a third party service needs a different
    format, it has to be hard-coded here or we need other changes to the offer model.

    :param offer: one offer (formerly extra)
    :type offer: ~integreat_cms.cms.models.offers.offer_template.OfferTemplate

    :param region: current region object
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :return: The post data of the offer's url
    :rtype: dict
    """
    post_data = offer.post_data
    if offer.use_postal_code == postal_code.POST:
        post_data.update({"search-plz": region.postal_code})
    if not post_data:
        return None
    return post_data


def transform_offer(offer, region):
    """
    Function to create a JSON from a single offer Object.

    :param offer: one offer (formerly extra)
    :type offer: ~integreat_cms.cms.models.offers.offer_template.OfferTemplate

    :param region: current region object
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :return: data necessary for API
    :rtype: dict
    """
    return {
        "name": offer.name,
        "alias": offer.slug,
        "url": get_url(offer, region),
        "post": get_post_data(offer, region),
        "thumbnail": offer.thumbnail,
    }


@json_response
# pylint: disable=unused-argument
def offers(request, region_slug, language_slug=None):
    """
    Function to iterate through all offers related to a region and adds them to a JSON.

    :param request: Django request
    :type request: ~django.http.HttpRequest
    :param region_slug: slug of a region
    :type region_slug: str
    :param language_slug: language slug
    :type language_slug: str

    :return: JSON object according to APIv3 offers endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    region = request.region
    result = [transform_offer(offer, region) for offer in region.offers.all()]
    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays



from ..custom_model_form import CustomModelForm
from ...models import PushNotification


class PushNotificationForm(CustomModelForm):
    """
    Form for creating and modifying push notification objects
    """

    class Meta:
        model = PushNotification
        fields = ["channel", "mode"]



import filecmp
import io
import zipfile

from os import listdir
from os.path import isfile, join

import pytest

from lxml.html import fromstring
from django.urls import reverse

from integreat_cms.cms.models import Page

from ..conftest import (
    ANONYMOUS,
    STAFF_ROLES,
    PRIV_STAFF_ROLES,
    MANAGEMENT,
    EDITOR,
    AUTHOR,
)


# pylint: disable=too-many-locals
@pytest.mark.django_db
# Override urls to serve XLIFF files
@pytest.mark.urls("tests.xliff.dummy_django_app.static_urls")
@pytest.mark.parametrize("xliff_version", ["xliff-1.2", "xliff-2.0"])
@pytest.mark.parametrize(
    "view,directory",
    [("download_xliff", "latest"), ("download_xliff_only_public", "only_public")],
)
def test_xliff_export(
    login_role_user, settings, tmp_path, xliff_version, view, directory
):
    """
    This test checks whether the xliff export works as expected

    :param login_role_user: The fixture providing the http client and the current role (see :meth:`~tests.conftest.login_role_user`)
    :type login_role_user: tuple

    :param settings: The fixture providing the django settings
    :type settings: :fixture:`settings`

    :param tmp_path: The fixture providing the directory for temporary files for this test case
    :type tmp_path: :fixture:`tmp_path`

    :param xliff_version: The XLIFF version to be tested
    :type xliff_version: str

    :param view: The name of the view for the export
    :type view: str

    :param directory: The directory that contains the expected XLIFF files
    :type directory: str
    """
    # Override used XLIFF version
    settings.XLIFF_EXPORT_VERSION = xliff_version
    client, role = login_role_user
    export_xliff = reverse(
        view, kwargs={"region_slug": "augsburg", "language_slug": "en"}
    )
    response = client.post(
        export_xliff, data={"selected_ids[]": [1, 2, 3, 4, 5, 14, 15]}
    )
    print(response.headers)
    if role in STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]:
        # If the role should be allowed to access the view, we expect a successful result
        assert response.status_code == 302
        page_tree = reverse(
            "pages", kwargs={"region_slug": "augsburg", "language_slug": "en"}
        )
        assert response.headers.get("Location") == page_tree
        response = client.get(page_tree)
        print(response.headers)
        # The download link to the xliff file is contained in the message
        parsed_content = fromstring(response.content.decode("utf-8"))
        download_links = [
            link.attrib["href"]
            for link in parsed_content.iter("a")
            if "data-auto-download" in link.attrib
        ]
        print(download_links)
        assert len(download_links) == 1
        response = client.get(download_links[0])
        print(response.headers)
        print(response.status_code)
        assert response.status_code == 200
        # Check if zip can be unzipped and contains the correct xliff files
        file = io.BytesIO(response.getvalue())
        expected_result_dir = f"tests/xliff/files/export/{xliff_version}/{directory}"
        with zipfile.ZipFile(file, "r") as zipped_file:
            assert zipped_file.testzip() is None
            assert set(zipped_file.namelist()) == {
                filename
                for filename in listdir(expected_result_dir)
                if isfile(join(expected_result_dir, filename))
            }
            zipped_file.extractall(path=tmp_path)
            for xliff_file in zipped_file.namelist():
                assert filecmp.cmp(
                    f"{tmp_path}/{xliff_file}", f"{expected_result_dir}/{xliff_file}"
                )
        # Check if existing translations are now "currently in translation"
        for page in Page.objects.filter(id__in=[1, 2]):
            translation = page.get_translation("en")
            if translation:
                assert translation.currently_in_translation
    elif role == ANONYMOUS:
        # For anonymous users, we want to redirect to the login form instead of showing an error
        assert response.status_code == 302
        assert (
            response.headers.get("location")
            == f"{settings.LOGIN_URL}?next={export_xliff}"
        )
    else:
        # For logged in users, we want to show an error if they get a permission denied
        assert response.status_code == 403


# pylint: disable=too-many-locals
@pytest.mark.django_db
def test_xliff_import(login_role_user, settings):
    """
    This test checks whether the xliff import works as expected

    :param login_role_user: The fixture providing the http client and the current role (see :meth:`~tests.conftest.login_role_user`)
    :type login_role_user: tuple

    :param settings: The fixture providing the django settings
    :type settings: :fixture:`settings`
    """
    # Test for english messages
    settings.LANGUAGE_CODE = "en"
    client, role = login_role_user
    upload_xliff = reverse(
        "upload_xliff", kwargs={"region_slug": "augsburg", "language_slug": "en"}
    )
    import_path = "tests/xliff/files/import"
    file_1 = "augsburg_de_en_1_2_willkommen.xliff"
    file_2 = "augsburg_de_en_2_1_willkommen-in-augsburg.xliff"
    # Upload once for successful import
    with open(f"{import_path}/{file_1}", encoding="utf-8") as f1:
        with open(f"{import_path}/{file_2}", encoding="utf-8") as f2:
            response1 = client.post(
                upload_xliff, data={"xliff_file": [f1, f2]}, format="multipart"
            )
            print(response1.headers)
    # Upload second time to test unchanged import
    with open(f"{import_path}/{file_1}", encoding="utf-8") as f1:
        with open(f"{import_path}/{file_2}", encoding="utf-8") as f2:
            response2 = client.post(
                upload_xliff, data={"xliff_file": [f1, f2]}, format="multipart"
            )
            print(response2.headers)

    if role in PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]:
        # Perform test twice to check whether unchanged diffs can be imported as well
        for response, msg in [
            (response1, "successfully"),
            (response2, "without changes"),
        ]:
            # If the role should be allowed to access the view, we expect a successful result
            assert response.status_code == 302
            page_tree = reverse(
                "pages", kwargs={"region_slug": "augsburg", "language_slug": "en"}
            )
            redirect_location = response.headers.get("Location")
            # If errors occur, we get redirected to the page tree
            assert redirect_location != page_tree
            # Check if xliff import view is correctly rendered
            response = client.get(redirect_location)
            print(response.headers)
            assert response.status_code == 200
            assert file_1 in response.content.decode("utf-8")
            assert file_2 in response.content.decode("utf-8")
            # Check if XLIFF import is correctly confirmed
            response = client.post(redirect_location)
            print(response.headers)
            assert response.status_code == 302
            # If the import has been successfully confirmed, we get redirected to the page tree
            assert response.headers.get("Location") == page_tree
            # Check if xliff import view is correctly rendered
            response = client.get(page_tree)
            print(response.headers)
            assert response.status_code == 200
            # Check if existing translations are now updated
            for page in Page.objects.filter(id__in=[1, 2]):
                translation = page.get_translation("en")
                assert translation.title == "Updated title"
                assert translation.content == "<p>Updated content</p>"
                assert not translation.currently_in_translation
                assert (
                    f'Page "{translation.title}" was imported {msg}.'
                    in response.content.decode("utf-8")
                )
                if translation.version > 1:
                    # If a translation already exists for this version, assert that the status is inherited
                    previous_translation = page.translations.get(
                        language__slug="en", version=translation.version - 1
                    )
                    assert previous_translation.status == translation.status
                else:
                    # Else, the status should be inherited from the source translation
                    assert translation.source_translation.status == translation.status
    elif role == ANONYMOUS:
        for response in [response1, response2]:
            # For anonymous users, we want to redirect to the login form instead of showing an error
            assert response.status_code == 302
            assert (
                response.headers.get("location")
                == f"{settings.LOGIN_URL}?next={upload_xliff}"
            )
    else:
        for response in [response1, response2]:
            # For logged in users, we want to show an error if they get a permission denied
            assert response.status_code == 403



import logging

from django.views.generic import View
from django.http import JsonResponse

from ...utils.linkcheck_utils import get_url_count

logger = logging.getLogger(__name__)


class LinkcheckStatsView(View):
    """
    Return the linkcheck counter stats
    """

    # pylint: disable=unused-argument
    def get(self, request, *args, **kwargs):
        r"""
        Retrieve the stats about valid/invalid/unchecked/ignored links

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """
        # Link count
        count_dict = get_url_count(kwargs.get("region_slug"))
        return JsonResponse(count_dict)



"""
This module contains all string representations of all day filter options, used by :class:`~integreat_cms.cms.forms.events.event_filter_form.EventFilterForm` and
:class:`~integreat_cms.cms.views.events.event_list_view.EventListView`.

The module also contains a constant :attr:`~integreat_cms.cms.constants.all_day.DATATYPE`, which contains the type of the constant
values linked to the strings and is used for correctly instantiating :class:`django.forms.TypedMultipleChoiceField`
instances in :class:`~integreat_cms.cms.forms.events.event_filter_form.EventFilterForm`.
"""
from django.utils.translation import ugettext_lazy as _


#: Only events which are all day long
READ = "READ"
#: Exclude events which are all day long
UNREAD = "UNREAD"

#: Choices to use these constants in a database field
CHOICES = [
    (READ, _("Read")),
    (UNREAD, _("Unread")),
]

INITIAL = [key for (key, val) in CHOICES]



"""
This is a collection of tags and filters for :class:`~integreat_cms.cms.models.push_notifications.push_notification.PushNotification`
objects.
"""
from django import template

register = template.Library()


@register.simple_tag
def get_translation(push_notification, language_slug):
    """
    This tag returns the most recent translation of the requested push notification in the requested language.

    :param push_notification: The requested push notification
    :type push_notification: ~integreat_cms.cms.models.push_notifications.push_notification.PushNotification

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :return: The push notification translation
    :rtype: ~integreat_cms.cms.models.push_notifications.push_notification_translation.PushNotificationTranslation
    """
    return push_notification.translations.filter(language__slug=language_slug).first()



import logging

from ...models import ImprintPageTranslation
from ..custom_content_model_form import CustomContentModelForm


logger = logging.getLogger(__name__)


class ImprintTranslationForm(CustomContentModelForm):
    """
    Form for creating and modifying imprint translation objects
    """

    class Meta:
        model = ImprintPageTranslation
        fields = ["title", "status", "content", "minor_edit"]



import logging

from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.shortcuts import render, redirect, get_object_or_404
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView

from ...constants import status
from ...decorators import permission_required
from ...models import Language
from .page_context_mixin import PageContextMixin

logger = logging.getLogger(__name__)


@method_decorator(permission_required("cms.view_page"), name="dispatch")
class PageRevisionView(PageContextMixin, TemplateView):
    """
    View for browsing the page revisions and restoring old page revisions
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "pages/page_revisions.html"
    #: The context dict passed to the template (see :class:`~django.views.generic.base.ContextMixin`)
    extra_context = {"current_menu_item": "pages"}

    def get(self, request, *args, **kwargs):
        r"""
        Render page revision slider

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.core.exceptions.PermissionDenied: If user does not have the permission to edit the specific page

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region
        page = get_object_or_404(region.pages, id=kwargs.get("page_id"))

        language = region.get_language_or_404(
            kwargs.get("language_slug"), only_active=True
        )

        page_translations = page.translations.filter(language=language)

        if not page_translations.exists():
            return redirect(
                "edit_page",
                **{
                    "page_id": page.id,
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )

        if not request.user.has_perm("cms.change_page_object", page):
            messages.warning(
                request,
                _("You don't have the permission to restore versions of this page."),
            )

        selected_revision = page_translations.filter(
            version=kwargs.get("selected_revision", page_translations.count())
        )

        if not selected_revision.exists():
            messages.error(request, _("This revision does not exist."))
            return redirect(
                "page_revisions",
                **{
                    "page_id": page.id,
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )

        if page.archived:
            messages.warning(
                request,
                _("You cannot restore versions of this page because it is archived."),
            )

        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "page": page,
                "page_translations": page_translations,
                "api_revision": page_translations.filter(status=status.PUBLIC).first(),
                "selected_revision": selected_revision.first(),
                "language": language,
            },
        )

    # pylint: disable=unused-argument, too-many-branches
    def post(self, request, *args, **kwargs):
        r"""
        Restore a previous revision of a page translation

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.core.exceptions.PermissionDenied: If user does not have the permission to edit the specific page

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region
        page = region.pages.get(id=kwargs.get("page_id"))

        language = Language.objects.get(slug=kwargs.get("language_slug"))

        revision = page.translations.filter(
            language=language, version=request.POST.get("revision")
        ).first()

        redirect_to_page_revisions = redirect(
            "page_revisions",
            **{
                "page_id": page.id,
                "region_slug": region.slug,
                "language_slug": language.slug,
            },
        )

        if not revision:
            messages.error(request, _("This revision does not exist."))
            return redirect_to_page_revisions

        if not request.user.has_perm("cms.change_page_object", page):
            raise PermissionDenied(
                f"{request.user!r} does not have the permission to restore {revision!r} of {page!r}"
            )

        current_revision = page.get_translation(language.slug)

        desired_status = request.POST.get("status")

        if not desired_status:
            # If the current version should be rejected, return to the latest version that is neither an auto save nor in review
            revision = page.translations.filter(
                language=language, status__in=[status.DRAFT, status.PUBLIC]
            ).first()
            if not revision:
                messages.error(
                    request,
                    _("You cannot reject changes if there is no version to return to."),
                )
                return redirect_to_page_revisions
            desired_status = revision.status
        elif desired_status not in dict(status.CHOICES):
            raise PermissionDenied(
                f"{request.user!r} tried to restore {revision!r} of {page!r} with invalid status {desired_status!r}"
            )

        # Assume that changing to an older revision is not a minor change by default
        minor_edit = False
        if (
            revision.slug == current_revision.slug
            and revision.title == current_revision.title
            and revision.content == current_revision.content
        ):
            minor_edit = True
            if desired_status == status.PUBLIC:
                if current_revision.status == status.PUBLIC:
                    messages.info(request, _("No changes detected, but date refreshed"))
                else:
                    messages.success(
                        request,
                        _("No changes detected, but status changed to published"),
                    )
            else:
                messages.error(
                    request,
                    _(
                        "This version is identical to the current version of this translation."
                    ),
                )
                return redirect_to_page_revisions

        if desired_status == status.DRAFT and not request.user.has_perm(
            "cms.publish_page_object", page
        ):
            raise PermissionDenied(
                f"{request.user!r} does not have the permission to restore {revision!r} of {page!r} as draft"
            )
        if desired_status == status.PUBLIC and not request.user.has_perm(
            "cms.publish_page_object", page
        ):
            raise PermissionDenied(
                f"{request.user!r} does not have the permission to restore the public {revision!r} of {page!r}"
            )

        # Delete all now outdated links
        current_revision.links.all().delete()
        # Create new version
        revision.pk = None
        revision.version = current_revision.version + 1
        revision.status = desired_status
        revision.minor_edit = minor_edit
        # Reset author to current user
        revision.creator = request.user
        revision.save()

        messages.success(request, _("The version was successfully restored"))

        return redirect_to_page_revisions



import json
import logging

from django.http import JsonResponse
from django.views.decorators.http import require_POST

from ...utils.content_edit_lock import lock_content, unlock_content, get_locking_user

logger = logging.getLogger(__name__)


@require_POST
# pylint: disable=unused-argument
def content_edit_lock_heartbeat(request, region_slug=None):
    """
    This function handles heartbeat requests.
    When a heartbeat is received, this function tries to extend the lock for a user
    who is editing some content.

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param region_slug: The slug of the current region, unused
    :type region_slug: str

    :return: Json object containing `success: true` if the lock could be acquired
    :rtype: ~django.http.JsonResponse
    """
    body = json.loads(request.body.decode("utf-8"))
    id_, type_ = json.loads(body["key"])
    force_take_over = body["force"]

    if force_take_over:
        locking_user = get_locking_user(id_, type_)
        if locking_user:
            logger.debug(
                "User %r took control over %s with id %s from %r",
                request.user,
                type_,
                id_,
                locking_user,
            )
            unlock_content(id_, type_, locking_user)

    success = lock_content(id_, type_, request.user)
    if not success:
        # If another user has locked the content before and a takeover was not forced, return the new locking user
        locking_user = get_locking_user(id_, type_)
    else:
        locking_user = request.user
    return JsonResponse(
        {"success": success, "lockingUser": locking_user.full_user_name}
    )


@require_POST
# pylint: disable=unused-argument
def content_edit_lock_release(request, region_slug=None):
    """
    This function handles unlock requests

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param region_slug: The slug of the current region, unused
    :type region_slug: str

    :return: Json object containing `success: true` if the content object could be unlocked
    :rtype: ~django.http.JsonResponse
    """
    body = json.loads(request.body.decode("utf-8"))
    id_, type_ = body

    success = unlock_content(id_, type_, request.user)
    return JsonResponse({"success": success})



"""
Contains a collection of tags for working with urls.
"""

from urllib.parse import urlparse, parse_qs, urlencode, urlunparse

from django import template


register = template.Library()


@register.simple_tag
def add_queries(url, *queries):
    r"""
    This filter adds a query to an url

    :param url: The url to modify
    :type url: str

    :param \*queries: The queries to add to the url. Should contain key-value pairs
    :type \*queries: tuple

    :return: The url with a modified query part
    :rtype: str
    """
    url = urlparse(url)
    url_query = parse_qs(url.query, keep_blank_values=True)
    while queries:
        key, value, *queries = queries
        url_query[key] = str(value)

    url = url._replace(query=urlencode(url_query, doseq=True))
    return urlunparse(url)



"""
This package contains all views related to statistics
"""
from .statistics_view import AnalyticsView
from .statistics_actions import (
    get_total_visits_ajax,
    get_visits_per_language_ajax,
)



"""
This module includes functions that are used as decorators in the API endpoints.
"""
import json
import threading
import logging
import random
import re

from functools import wraps

from urllib import request, parse, error

from django.http import JsonResponse, Http404
from django.views.decorators.csrf import csrf_exempt
from django.conf import settings

from ..cms.models import Region, Language
from ..cms.constants import feedback_ratings

logger = logging.getLogger(__name__)


def feedback_handler(func):
    """
    Decorator definition for feedback API functions and methods

    :param func: decorated function
    :type func: ~collections.abc.Callable

    :return: The decorated feedback view function
    :rtype: ~collections.abc.Callable
    """

    @csrf_exempt
    @wraps(func)
    def handle_feedback(request, region_slug, language_slug):
        """
        Parse feedback API request parameters

        :param request: Django request
        :type request: ~django.http.HttpRequest

        :param region_slug: slug of a region
        :type region_slug: str

        :param language_slug: slug of a language
        :type language_slug: str

        :return: The decorated feedback view function
        :rtype: ~collections.abc.Callable
        """
        if request.method != "POST":
            return JsonResponse({"error": "Invalid request."}, status=405)
        try:
            region = Region.objects.get(slug=region_slug)
            language = Language.objects.get(slug=language_slug)
        except Region.DoesNotExist:
            return JsonResponse(
                {"error": f'No region found with slug "{region_slug}"'}, status=404
            )
        except Language.DoesNotExist:
            return JsonResponse(
                {"error": f'No language found with slug "{language_slug}"'}, status=404
            )
        if request.content_type == "application/json":
            data = json.loads(request.body.decode())
        else:
            data = request.POST.dict()
        comment = data.pop("comment", "")
        rating = data.pop("rating", None)
        category = data.pop("category", None)

        if rating not in [None, "up", "down"]:
            return JsonResponse({"error": "Invalid rating."}, status=400)
        if comment == "" and not rating:
            return JsonResponse(
                {"error": "Either comment or rating is required."}, status=400
            )
        if rating == "up":
            rating_normalized = feedback_ratings.POSITIVE
        elif rating == "down":
            rating_normalized = feedback_ratings.NEGATIVE
        else:
            rating_normalized = feedback_ratings.NOT_STATED
        is_technical = category == "Technisches Feedback"
        return func(data, region, language, comment, rating_normalized, is_technical)

    return handle_feedback


def json_response(function):
    """
    This decorator can be used to catch :class:`~django.http.Http404` exceptions and convert them to a :class:`~django.http.JsonResponse`.
    Without this decorator, the exceptions would be converted to :class:`~django.http.HttpResponse`.

    :param function: The view function which should always return JSON
    :type function: ~collections.abc.Callable

    :return: The decorated function
    :rtype: ~collections.abc.Callable
    """

    @wraps(function)
    def wrap(request, *args, **kwargs):
        r"""
        The inner function for this decorator.
        It tries to execute the decorated view function and returns the unaltered result with the exception of a
        :class:`~django.http.Http404` error, which is converted into JSON format.

        :param request: Django request
        :type request: ~django.http.HttpRequest

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied kwargs
        :type \**kwargs: dict

        :return: The response of the given function or an 404 :class:`~django.http.JsonResponse`
        :rtype: ~django.http.JsonResponse
        """
        try:
            return function(request, *args, **kwargs)
        except Http404 as e:
            return JsonResponse({"error": str(e) or "Not found."}, status=404)

    return wrap


def matomo_tracking(func):
    """
    This decorator is supposed to be applied to API content endpoints. It will track
    the request in Matomo. The request to the Matomo API is executed asynchronously in its
    own thread to not block the Integreat CMS API request.

    Only the URL and the User Agent will be sent to Matomo.

    :param func: decorated function
    :type func: ~collections.abc.Callable

    :return: The decorated feedback view function
    :rtype: ~collections.abc.Callable
    """

    @wraps(func)
    def wrap(request, *args, **kwargs):
        r"""
        The inner function for this decorator.

        :param request: Django request
        :type request: ~django.http.HttpRequest

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied kwargs
        :type \**kwargs: dict

        :return: The response of the given function or an 404 :class:`~django.http.JsonResponse`
        :rtype: ~django.http.JsonResponse
        """
        if (
            not settings.MATOMO_TRACKING
            or not request.region.matomo_id
            or not request.region.matomo_token
        ):
            return func(request, *args, **kwargs)
        data = {
            "idsite": request.region.matomo_id,
            "token_auth": request.region.matomo_token,
            "rec": 1,
            "url": request.build_absolute_uri(),
            "urlref": settings.BASE_URL,
            "ua": request.META.get("HTTP_USER_AGENT", "unknown user agent"),
            "cip": f"{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}",
        }

        t = threading.Thread(target=matomo_request, args=(settings.MATOMO_URL, data))
        t.setDaemon(True)
        t.start()
        return func(request, *args, **kwargs)

    def matomo_request(host, data):
        """
        Wrap HTTP request to Matomo in threaded function.

        :param host: Hostname + protocol of Matomo installation
        :type host:

        :param data: Data to send to Matomo API
        :type data: dict
        """
        data = parse.urlencode(data)
        url = f"{host}/matomo.php?{data}"
        req = request.Request(url)
        try:
            with request.urlopen(req):
                pass
        except error.HTTPError as e:
            logger.error(
                "Matomo Tracking API request to %r failed with: %s",
                # Mask auth token in log
                re.sub(r"&token_auth=[^&]+", "&token_auth=********", url),
                e,
            )

    return wrap



"""
Configuration of GVZ API app
"""
import os
import sys
import logging
import json
import requests
from django.apps import AppConfig
from django.conf import settings

logger = logging.getLogger(__name__)


class GvzApiConfig(AppConfig):
    """
    GVZ API config inheriting the django AppConfig
    """

    name = "integreat_cms.gvz_api"
    api_available = False

    def ready(self):
        """
        Checking if API is available
        """
        # Only check availability if running a server
        if "runserver" in sys.argv or "APACHE_PID_FILE" in os.environ:
            if settings.GVZ_API_ENABLED:
                try:
                    response = requests.get(f"{settings.GVZ_API_URL}/api/", timeout=3)
                    # Require the response to return a valid JSON, otherwise it's probably an error
                    assert json.loads(response.text)
                    logger.info("GVZ API is available at: %r", settings.GVZ_API_URL)
                    self.api_available = True
                except (
                    json.decoder.JSONDecodeError,
                    requests.exceptions.RequestException,
                    requests.exceptions.Timeout,
                    AssertionError,
                ) as e:
                    logger.exception(e)
                    logger.error(
                        "GVZ API is unavailable. You won't be able to "
                        "automatically import region coordinates and aliases."
                    )
            else:
                logger.info("GVZ API is disabled")



import logging
import os
from datetime import datetime


import magic

from django import forms
from django.conf import settings
from django.utils.translation import ugettext as _

from ...constants import allowed_media
from ...models import MediaFile
from ...utils.media_utils import generate_thumbnail
from ..custom_model_form import CustomModelForm

logger = logging.getLogger(__name__)


class ReplaceMediaFileForm(CustomModelForm):
    """
    Form for replacing media file objects
    """

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = MediaFile
        #: The fields of the model which should be handled by this form
        fields = ("file", "name", "thumbnail", "file_size", "last_modified")

    def __init__(self, data=None, files=None, instance=None):
        """
        Initialize UploadMediaFileForm form

        :param data: A dictionary-like object containing all given HTTP POST parameters
        :type data: django.http.QueryDict

        :param files: A dictionary-like object containing all uploaded files
        :type files: django.utils.datastructures.MultiValueDict

        :param instance: This form's instance
        :type instance: ~integreat_cms.cms.models.media.media_file.MediaFile
        """

        # instantiate ModelForm
        super().__init__(data=data, files=files, instance=instance)

        # Make fields not required because they're filled automatically
        self.fields["name"].required = False
        self.fields["thumbnail"].required = False
        self.fields["file_size"].required = False
        self.fields["last_modified"].required = False

        self.original_file_path = instance.file.path
        self.original_thumbnail_path = (
            instance.thumbnail.path if instance.thumbnail else None
        )

    def clean(self):
        """
        Validate form fields which depend on each other, see :meth:`django.forms.Form.clean`:
        Check whether a file was uploaded and whether it's type matches the original file's type.
        If the file type is invalid, add a :class:`~django.core.exceptions.ValidationError`.

        :return: The cleaned form data
        :rtype: dict
        """
        cleaned_data = super().clean()

        file = cleaned_data.get("file")

        # Only validate type if a file is uploaded - otherwise, the form throws an error anyways
        if file:
            # Check magic bytes for actual file type - the content_type of the file can easily be forged
            new_type = magic.from_buffer(file.read(), mime=True)
            if new_type != self.instance.type:
                # Check whether we have a more readable version of the mime type
                new_type_display = dict(allowed_media.CHOICES).get(new_type, new_type)
                self.add_error(
                    "file",
                    forms.ValidationError(
                        _(
                            "The file type of the new file ({}) does not match the original file's type ({})."
                        ).format(new_type_display, self.instance.get_type_display())
                    ),
                )

            # Set the visible name in the database to the filename of the new file
            cleaned_data["name"] = file.name

        # If everything looks good until now, generate a thumbnail and an optimized image
        if not self.errors and self.instance.type.startswith("image"):
            optimized_image = generate_thumbnail(
                file, settings.MEDIA_OPTIMIZED_SIZE, False
            )
            if not optimized_image:
                self.add_error(
                    "file",
                    forms.ValidationError(
                        _("This image file is corrupt."), code="invalid"
                    ),
                )
            else:
                cleaned_data["file"] = optimized_image
                cleaned_data["thumbnail"] = generate_thumbnail(file)

        # Add the calculated file_size to the form data
        if cleaned_data.get("file"):
            cleaned_data["file_size"] = cleaned_data.get("file").size
        cleaned_data["last_modified"] = datetime.now()

        logger.debug(
            "ReplaceMediaFileForm validated [2] with cleaned data %r", cleaned_data
        )
        return cleaned_data

    def save(self, commit=True):
        # Remove old file
        os.remove(self.original_file_path)
        logger.debug("Removed old file %r", self.original_file_path)

        # Remove old thumbnail
        if self.original_thumbnail_path:
            os.remove(self.original_thumbnail_path)
            logger.debug("Removed old thumbnail %r", self.original_thumbnail_path)

        super().save(commit=commit)



"""
This module contains colors which aim to be good distinguishable by humans.
They are used to display graphs with overlapping lines (e.g. statistics).
"""

#: The default color (blue-500 from tailwind css).
DEFAULT = "#3b82f6"

#: A list of 20 distinguishable colors (taken from https://sashamaps.net/docs/resources/20-colors/)
CHOICES = [
    "#e6194b",
    "#000075",
    "#ffe119",
    "#f58231",
    "#3cb44b",
    "#46f0f0",
    "#f032e6",
    "#800000",
    "#fffac8",
    "#aaffc3",
    "#bcf60c",
    "#4363d8",
    "#008080",
    "#9a6324",
    "#911eb4",
    "#808000",
    "#e6beff",
    "#808080",
    "#fabebe",
    "#ffd8b1",
]



"""
APIv3 endpoint for the events (list)
"""
from django.http import JsonResponse

from ....cms.models import EventListFeedback
from ...decorators import json_response, feedback_handler


@feedback_handler
@json_response
# pylint: disable=unused-argument
def event_list_feedback(data, region, language, comment, rating, is_technical):
    """
    Store feedback on events list in database

    :param data: HTTP request body data
    :type data: dict

    :param region: The region of this sitemap's urls
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The language of this sitemap's urls
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param comment: The comment sent as feedback
    :type comment: str

    :param rating: up or downvote, neutral
    :type rating: str

    :param is_technical: is feedback on content or on tech
    :type is_technical: bool

    :return: JSON object according to APIv3 event list feedback endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    EventListFeedback.objects.create(
        region=region,
        language=language,
        rating=rating,
        comment=comment,
        is_technical=is_technical,
    )
    return JsonResponse({"success": "Feedback successfully submitted"}, status=201)



"""
This module contains view actions for region user objects.
"""
import logging

from django.contrib import messages
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from ...decorators import permission_required
from ...utils.welcome_mail_utils import send_welcome_mail

logger = logging.getLogger(__name__)


@require_POST
@permission_required("cms.delete_user")
# pylint: disable=unused-argument
def delete_region_user(request, region_slug, user_id):
    """
    This view deletes a region user

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param region_slug: The slug of the current region
    :type region_slug: str

    :param user_id: The id of the user which should be deleted
    :type user_id: int

    :return: A redirection to region user list
    :rtype: ~django.http.HttpResponseRedirect
    """

    region = request.region
    user = get_object_or_404(region.region_users, id=user_id)
    if user.regions.count() == 1:
        logger.info("%r deleted %r", request.user, user)
        user.delete()
        messages.success(request, _("User {} was successfully deleted.").format(user))
    else:
        user.regions.remove(region)
        user.save()
        logger.info(
            "%r removed %r from %r",
            request.user,
            user,
            region,
        )
        messages.success(
            request,
            _("User {} was successfully removed from this region.").format(user),
        )

    return redirect("region_users", region_slug=region.slug)


@require_POST
@permission_required("cms.change_user")
# pylint: disable=unused-argument
def resend_activation_link_region(request, region_slug, user_id):
    """
    Resends an activation link to a region user

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param region_slug: The slug of the current region
    :type region_slug: str

    :param user_id: users id to send the activation link
    :type user_id: int

    :return: A redirection to region user list
    :rtype: ~django.http.HttpResponseRedirect
    """
    region = request.region
    user = get_object_or_404(region.region_users, id=user_id)
    send_welcome_mail(request, user, activation=True)
    return redirect("region_users", region_slug=region.slug)



import json
import datetime
import logging
from webauthn import verify_authentication_response, base64url_to_bytes
from webauthn.helpers.structs import AuthenticationCredential
from webauthn.helpers.exceptions import InvalidAuthenticationResponse

from django.conf import settings
from django.contrib.auth import get_user_model, login as auth_login
from django.utils.translation import ugettext as _
from django.http import JsonResponse
from django.views.generic import View

logger = logging.getLogger(__name__)


class MfaVerifyView(View):
    """
    Verify the response to the challenge generated in :class:`~integreat_cms.cms.views.authentication.mfa.mfa_assert_view.MfaAssertView`.
    After a successful verification, the user is logged in.
    """

    def post(self, request):
        """

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :return: The mfa challenge as JSON
        :rtype: ~django.http.JsonResponse
        """

        if "mfa_user_id" not in request.session:
            return JsonResponse(
                {"success": False, "error": _("You need to log in first")}, status=403
            )

        user = get_user_model().objects.get(id=request.session["mfa_user_id"])

        challenge = request.session["challenge"]
        assertion_response = json.loads(request.body)
        credential_id = assertion_response["id"]

        key = user.mfa_keys.get(key_id=base64url_to_bytes(credential_id))

        try:
            authentication_verification = verify_authentication_response(
                credential=AuthenticationCredential.parse_raw(request.body),
                expected_challenge=base64url_to_bytes(challenge),
                expected_rp_id=settings.HOSTNAME,
                expected_origin=settings.BASE_URL,
                credential_public_key=key.public_key,
                credential_current_sign_count=key.sign_count,
            )
        except InvalidAuthenticationResponse as e:
            logger.exception(e)
            return JsonResponse(
                {"success": False, "error": "Authentication rejected"}, status=403
            )

        # Update counter.
        key.sign_count = authentication_verification.new_sign_count
        key.last_usage = datetime.datetime.now()
        key.save()

        auth_login(request, user, backend="django.contrib.auth.backends.ModelBackend")

        return JsonResponse({"success": True})



import logging

from django.urls import reverse

from ...forms import LanguageTreeNodeForm
from ..form_views import CustomCreateView

logger = logging.getLogger(__name__)


class LanguageTreeNodeCreateView(CustomCreateView):
    """
    Class that handles creating language tree nodes.
    This view is available within regions.
    """

    #: The form class of this form view
    form_class = LanguageTreeNodeForm

    def get_success_url(self):
        """
        Determine the URL to redirect to when the form is successfully validated

        :return: The url to redirect on success
        :rtype: str
        """
        if "submit_add_new" in self.request.POST:
            # If the "Create and add more" button was clicked, redirect to the new creation form
            return reverse(
                "new_languagetreenode",
                kwargs={"region_slug": self.request.region.slug},
            )
        return super().get_success_url()

    def get_form_kwargs(self):
        """
        Return the keyword arguments for instantiating the form

        :return: The form kwargs
        :rtype: dict
        """
        kwargs = super().get_form_kwargs()
        kwargs["region"] = self.request.region
        return kwargs



import json
import logging

from copy import deepcopy

from zoneinfo import available_timezones
from django import forms
from django.conf import settings
from django.utils.translation import override, ugettext_lazy as _
from django.apps import apps

from ....nominatim_api.nominatim_api_client import NominatimApiClient
from ....gvz_api.utils import GvzRegion
from ...constants import status
from ...models import Region, Page, LanguageTreeNode
from ...utils.matomo_api_manager import MatomoException
from ...utils.slug_utils import generate_unique_slug_helper
from ...utils.translation_utils import ugettext_many_lazy as __
from ..icon_widget import IconWidget
from ..custom_model_form import CustomModelForm


logger = logging.getLogger(__name__)


def get_timezone_choices():
    """
    This method generates the options for the second timezone dropdown

    :return: A list of all available timezones
    :rtype: list
    """
    timezones = list(available_timezones())
    timezones.sort()
    return [("", "---------")] + [
        (tz, tz.split("/", 1)[1].replace("_", " ")) for tz in timezones if "/" in tz
    ]


def get_timezone_area_choices():
    """
    This method generates the options for the first timezone dropdown.
    It displays the general area of a country or city. Often the continent.

    :return: A list of the general areas of the timezones
    :rtype: list

    """
    timezone_regions = list(
        {
            tz.split("/")[0]
            for tz in available_timezones()
            if "/" in tz and "Etc" not in tz and "SystemV" not in tz
        }
    )
    timezone_regions.sort()
    return (
        [("", "---------")]
        + [(tzr, tzr) for tzr in timezone_regions]
        + [("Etc", _("Other timezones"))]
    )


class RegionForm(CustomModelForm):
    """
    Form for creating and modifying region objects
    """

    duplicated_region = forms.ModelChoiceField(
        queryset=Region.objects.all(),
        empty_label=_("Do no import initial content"),
        required=False,
    )

    timezone_area = forms.ChoiceField(
        choices=get_timezone_area_choices,
        label=_("Timezone area"),
        required=False,
    )

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = Region
        #: The fields of the model which should be handled by this form
        fields = [
            "name",
            "common_id",
            "slug",
            "events_enabled",
            "locations_enabled",
            "chat_enabled",
            "push_notifications_enabled",
            "latitude",
            "longitude",
            "longitude_min",
            "latitude_min",
            "longitude_max",
            "latitude_max",
            "postal_code",
            "admin_mail",
            "statistics_enabled",
            "matomo_id",
            "matomo_token",
            "status",
            "page_permissions_enabled",
            "administrative_division",
            "aliases",
            "icon",
            "administrative_division_included",
            "offers",
            "short_urls_enabled",
            "custom_prefix",
            "tunews_enabled",
            "timezone",
            "fallback_translations_enabled",
            "summ_ai_enabled",
        ]
        #: The widgets which are used in this form
        widgets = {
            "timezone": forms.Select(choices=get_timezone_choices()),
            "icon": IconWidget(),
            "offers": forms.CheckboxSelectMultiple(),
        }

    def __init__(self, *args, **kwargs):
        r"""
        Initialize region form

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict
        """
        super().__init__(*args, **kwargs)
        if self.instance and "/" in self.instance.timezone:
            self.fields["timezone_area"].initial = self.instance.timezone.split("/")[0]
        self.fields["slug"].required = False
        # Do not require coordinates because they might be automatically filled
        self.fields["latitude"].required = False
        self.fields["longitude"].required = False
        if not settings.SUMM_AI_ENABLED:
            self.fields["summ_ai_enabled"].disabled = True

    def save(self, commit=True):
        """
        This method extends the default ``save()``-method of the base :class:`~django.forms.ModelForm` to set attributes
        which are not directly determined by input fields.

        :param commit: Whether or not the changes should be written to the database
        :type commit: bool

        :return: The saved region object
        :rtype: ~integreat_cms.cms.models.regions.region.Region
        """

        # Only duplicate content if region is created and a region was selected
        duplicate_region = (
            not self.instance.id and self.cleaned_data["duplicated_region"]
        )

        # Save CustomModelForm
        region = super().save(commit=commit)

        if duplicate_region:
            source_region = self.cleaned_data["duplicated_region"]
            # Duplicate language tree
            logger.info("Duplicating language tree of %r to %r", source_region, region)
            duplicate_language_tree(source_region, region)
            # Duplicate pages
            logger.info("Duplicating page tree of %r to %r", source_region, region)
            duplicate_pages(source_region, region)
            # Duplicate Imprint
            logger.info("Duplicating imprint of %r to %r", source_region, region)
            duplicate_imprint(source_region, region)
            # Duplicate media content
            duplicate_media(source_region, region)

        return region

    def clean(self):
        """
        Validate form fields which depend on each other, see :meth:`django.forms.Form.clean`

        :return: The cleaned form data
        :rtype: dict
        """
        cleaned_data = super().clean()
        # Check wether statistics can be enabled
        if cleaned_data["statistics_enabled"] and not cleaned_data["matomo_token"]:
            self.add_error(
                "statistics_enabled",
                _(
                    "Statistics can only be enabled when a valid access token is supplied."
                ),
            )
        # Automatically set the Matomo ID
        if cleaned_data["matomo_token"]:
            try:
                cleaned_data["matomo_id"] = self.instance.statistics.get_matomo_id(
                    token_auth=cleaned_data["matomo_token"]
                )
            except MatomoException as e:
                logger.exception(e)
                self.add_error(
                    "matomo_token", _("The provided access token is invalid.")
                )
        else:
            cleaned_data["matomo_id"] = None

        # Get additional data from GVZ API
        if apps.get_app_config("gvz_api").api_available:
            gvz_region = GvzRegion(
                region_name=cleaned_data["name"],
                region_ags=cleaned_data["common_id"],
                region_type=cleaned_data["administrative_division"],
            )
            logger.debug("GVZ API match: %r", gvz_region)
            if gvz_region.child_coordinates and not cleaned_data.get("aliases"):
                cleaned_data["aliases"] = gvz_region.child_coordinates
            if gvz_region.longitude and not cleaned_data.get("longitude"):
                cleaned_data["longitude"] = gvz_region.longitude
            if gvz_region.latitude and not cleaned_data.get("latitude"):
                cleaned_data["latitude"] = gvz_region.latitude
            if gvz_region.ags and not cleaned_data.get("common_id"):
                cleaned_data["common_id"] = gvz_region.ags

        # If the coordinates could not be filled automatically and have not been filled manually either, throw an error
        if not cleaned_data.get("latitude"):
            self.add_error(
                "latitude",
                forms.ValidationError(
                    _(
                        "Could not retrieve the coordinates automatically, please fill the field manually."
                    ),
                    code="required",
                ),
            )
        if not cleaned_data.get("longitude"):
            self.add_error(
                "longitude",
                forms.ValidationError(
                    _(
                        "Could not retrieve the coordinates automatically, please fill the field manually."
                    ),
                    code="required",
                ),
            )

        # Automatically fill the bounding box coordinates
        cleaned_data = self.autofill_bounding_box(cleaned_data)

        logger.debug("RegionForm validated [2] with cleaned data %r", cleaned_data)
        return cleaned_data

    def clean_slug(self):
        """
        Validate the slug field (see :ref:`overriding-modelform-clean-method`)

        :return: A unique slug based on the input value
        :rtype: str
        """
        return generate_unique_slug_helper(self, "region")

    def clean_custom_prefix(self):
        """
        Validate the custom prefix field. (see :ref:`overriding-modelform-clean-method`)

        :return: The given prefix or ``None`` if it is invalid
        :rtype: str
        """
        cleaned_data = self.cleaned_data
        # Validate custom prefix
        if cleaned_data.get("custom_prefix"):
            # Get the administrative divisions as conflicting options
            administrative_divisions = [
                label
                for choice, label in self.fields["administrative_division"].choices
            ]
            for language_slug, language_name in settings.LANGUAGES:
                # Check if at least one translation of the labels matches the prefix
                with override(language_slug):
                    # Force evaluation of lazy-translated text
                    translated_administrative_divisions = list(
                        map(str, administrative_divisions)
                    )
                # Check if custom prefix could also be set via the administrative division
                if (
                    cleaned_data.get("custom_prefix")
                    in translated_administrative_divisions
                ):
                    error_messages = []
                    # Get currently selected administrative division
                    selected_administrative_division = dict(
                        self.fields["administrative_division"].choices
                    )[cleaned_data.get("administrative_division")]
                    # Check if administrative division needs to be changed to translated version
                    if cleaned_data.get("custom_prefix") in administrative_divisions:
                        desired_administrative_division = cleaned_data.get(
                            "custom_prefix"
                        )
                    else:
                        # Get index of translated administrative division
                        index = translated_administrative_divisions.index(
                            cleaned_data.get("custom_prefix")
                        )
                        # Get original label which needs to be selected in list
                        desired_administrative_division = administrative_divisions[
                            index
                        ]
                    if (
                        selected_administrative_division
                        == desired_administrative_division
                    ):
                        error_messages.append(
                            _(
                                "'{}' is already selected as administrative division."
                            ).format(selected_administrative_division)
                        )
                    else:
                        error_messages.append(
                            _("Please select '{}' as administrative division.").format(
                                desired_administrative_division
                            )
                        )
                    # Check if default language needs to be changed in order to use this administrative division
                    if (
                        not self.instance.default_language
                        or self.instance.default_language.native_name != language_name
                    ):
                        error_messages.append(
                            _(
                                "Please set {} as default language for this region."
                            ).format(_(language_name))
                        )
                    # Check if administrative division is included in name yet
                    if not cleaned_data.get("administrative_division_included"):
                        error_messages.append(
                            _("Please enable '{}'.").format(
                                self.fields["administrative_division_included"].label
                            )
                        )
                    self.add_error(
                        "custom_prefix",
                        __(*error_messages),
                    )
        # Check if administrative division is also included in the name and allow only one of both prefix options
        if cleaned_data.get("custom_prefix") and cleaned_data.get(
            "administrative_division_included"
        ):
            self.add_error(
                "custom_prefix",
                _(
                    "You cannot include the administrative division into the name and use a custom prefix at the same time."
                ),
            )
        return cleaned_data.get("custom_prefix")

    def clean_aliases(self):
        """
        Validate the aliases field (see :ref:`overriding-modelform-clean-method`).

        :return: The valid aliases
        :rtype: dict
        """
        cleaned_aliases = self.cleaned_data["aliases"]
        # If a string is given, try to load as JSON string
        if isinstance(cleaned_aliases, str):
            try:
                cleaned_aliases = json.loads(cleaned_aliases)
            except json.JSONDecodeError:
                self.add_error("aliases", _("Enter a valid JSON."))
        # Convert None to an empty dict
        return cleaned_aliases or {}

    def clean_summ_ai_enabled(self):
        """
        Validate the summ_ai_enabled field (see :ref:`overriding-modelform-clean-method`)

        :return: The validated field whether SUMM.AI is enabled
        :rtype: str
        """
        if self.cleaned_data.get("summ_ai_enabled") and not settings.SUMM_AI_ENABLED:
            self.add_error(
                "summ_ai_enabled", _("Currently SUMM.AI is globally deactivated")
            )
            return False
        return self.cleaned_data.get("summ_ai_enabled")

    @staticmethod
    def autofill_bounding_box(cleaned_data):
        """
        Automatically fill the bounding box coordinates

        :param cleaned_data: The partially cleaned data
        :type cleaned_data: dict

        :return: The updated cleaned data
        :rtype: dict
        """
        # When the Nominatim API is enabled, auto fill the bounding box coordinates
        if settings.NOMINATIM_API_ENABLED:
            nominatim_api_client = NominatimApiClient()
            bounding_box = nominatim_api_client.get_bounding_box(
                administrative_division=cleaned_data.get("administrative_division"),
                name=cleaned_data.get("name"),
                aliases=cleaned_data.get("aliases"),
            )
            if bounding_box:
                # Update bounding box values if not set manually
                if not cleaned_data.get("latitude_min"):
                    cleaned_data["latitude_min"] = bounding_box.latitude_min
                if not cleaned_data.get("latitude_max"):
                    cleaned_data["latitude_max"] = bounding_box.latitude_max
                if not cleaned_data.get("longitude_min"):
                    cleaned_data["longitude_min"] = bounding_box.longitude_min
                if not cleaned_data.get("longitude_max"):
                    cleaned_data["longitude_max"] = bounding_box.longitude_max

        return cleaned_data


def duplicate_language_tree(
    source_region,
    target_region,
    source_parent=None,
    target_parent=None,
    logging_prefix="",
):
    """
    Function to duplicate the language tree of one region to another.

    Usage: duplicate_language_tree(source_region, target_region)

    This is a recursive function to walk the whole language tree. It starts at root level with the default parent None.
    The recursion is necessary because the new nodes need their correct (also duplicated) parent node.

    :param source_region: The region from which the language tree should be duplicated
    :type source_region: ~integreat_cms.cms.models.regions.region.Region

    :param target_region: The region to which the language tree should be added
    :type target_region: ~integreat_cms.cms.models.regions.region.Region

    :param source_parent: The current parent node id of the recursion
    :type source_parent: ~integreat_cms.cms.models.languages.language_tree_node.LanguageTreeNode

    :param target_parent: The node of the target region which is the duplicate of the source parent node
    :type target_parent: ~integreat_cms.cms.models.pages.page.Page

    :param logging_prefix: recursion level to get a pretty log output
    :type logging_prefix: str
    """
    logger.debug(
        "%s Duplicating child nodes",
        logging_prefix + "└─",
    )
    logging_prefix += "  "

    # Iterate over all children of the current source parent, beginning with the root node
    source_nodes = source_region.language_tree_nodes.filter(parent=source_parent)
    num_source_nodes = len(source_nodes)
    for i, source_node in enumerate(source_nodes):
        logger.debug(
            "%s Duplicating node %r",
            logging_prefix + ("└─" if i == num_source_nodes - 1 else "├─"),
            source_node,
        )
        row_logging_prefix = logging_prefix + (
            "  " if i == num_source_nodes - 1 else "│  "
        )
        if target_parent:
            # If the target parent already exist, we inherit its tree id for all its sub nodes
            target_tree_id = target_parent.tree_id
        else:
            # If the language tree node is a root node, we need to assign a new tree id
            last_root_node = LanguageTreeNode.get_last_root_node()
            target_tree_id = last_root_node.tree_id + 1
            logger.debug(
                "%s Node is a root node, assigning new tree_id %r",
                row_logging_prefix + "├─",
                target_tree_id,
            )
        # Copy the target node to leave the source node unchanged for next iteration
        target_node = deepcopy(source_node)
        # Change the region and parent to its new values
        target_node.region = target_region
        target_node.parent = target_parent
        # Set new tree id
        target_node.tree_id = target_tree_id
        # Delete the primary key to force an insert
        target_node.pk = None
        # Check if the resulting node is valid
        target_node.full_clean()
        # Save the duplicated node
        target_node.save()
        logger.debug(
            "%s Created target node %r",
            row_logging_prefix + ("└─" if source_node.is_leaf() else "├─"),
            target_node,
        )
        if not source_node.is_leaf():
            # Call the function recursively for all children of the current node
            duplicate_language_tree(
                source_region,
                target_region,
                source_node,
                target_node,
                row_logging_prefix,
            )


def duplicate_pages(
    source_region,
    target_region,
    source_parent=None,
    target_parent=None,
    logging_prefix="",
):
    """
    Function to duplicate all pages of one region to another.

    Usage: duplicate_pages(source_region, target_region)

    This is a recursive function to walk the whole page tree. It starts at root level with the default parent None.
    The recursion is necessary because the new pages need their correct (also duplicated) parent page.

    :param source_region: The region from which the pages should be duplicated
    :type source_region: ~integreat_cms.cms.models.regions.region.Region

    :param target_region: The region to which the pages should be added
    :type target_region: ~integreat_cms.cms.models.regions.region.Region

    :param source_parent: The current parent page id of the recursion
    :type source_parent: ~integreat_cms.cms.models.pages.page.Page

    :param target_parent: The page of the target region which is the duplicate of the source parent page
    :type target_parent: ~integreat_cms.cms.models.pages.page.Page

    :param logging_prefix: recursion level to get a pretty log output
    :type logging_prefix: str
    """

    logger.debug(
        "%s Duplicating child nodes",
        logging_prefix + "└─",
    )
    logging_prefix += "  "

    # At first, get all pages from the source region with a specific parent page
    # As the parent will be None for the initial call, this returns all pages from the root level
    source_pages = source_region.pages.filter(parent=source_parent)
    num_source_pages = len(source_pages)
    for i, source_page in enumerate(source_pages):
        logger.debug(
            "%s Duplicating page %r",
            logging_prefix + ("└─" if i == num_source_pages - 1 else "├─"),
            source_page,
        )
        row_logging_prefix = logging_prefix + (
            "  " if i == num_source_pages - 1 else "│  "
        )
        if target_parent:
            # If the target parent already exist, we inherit its tree id for all its subpages
            target_tree_id = target_parent.tree_id
        else:
            # If the page is a root page, we need to assign a new tree id
            last_root_page = Page.get_last_root_node()
            target_tree_id = last_root_page.tree_id + 1
            logger.debug(
                "%s Page is a root page, assigning new tree_id %r",
                row_logging_prefix + "├─",
                target_tree_id,
            )
        # Copy the target node to leave the source node unchanged for next iteration
        target_page = deepcopy(source_page)
        # Set the parent of the new page to the previously created target parent
        target_page.parent = target_parent
        # Set the region of the new page to the target region
        target_page.region = target_region
        # Set new tree id
        target_page.tree_id = target_tree_id
        # Delete the primary key to duplicate the object instance instead of updating it
        target_page.pk = None
        # Set push API token to blank for duplicated page
        target_page.api_token = ""
        # Check if the page is valid
        target_page.full_clean()
        # Save duplicated page
        target_page.save()
        logger.debug(
            "%s Created %r",
            row_logging_prefix + "├─",
            target_page,
        )
        duplicate_page_translations(source_page, target_page, row_logging_prefix)
        if not source_page.is_leaf():
            # Recursively call this function with the current pages as new parents
            duplicate_pages(
                source_region,
                target_region,
                source_page,
                target_page,
                row_logging_prefix,
            )


def duplicate_page_translations(source_page, target_page, logging_prefix):
    """
    Duplicate all translations of a given source page to a given target page

    :param source_page: The given source page
    :type source_page: ~integreat_cms.cms.models.pages.page.Page

    :param target_page: The desired target page
    :type target_page: ~integreat_cms.cms.models.pages.page.Page

    :param logging_prefix: The prefix to be used for logging
    :type logging_prefix: str
    """
    logger.debug(
        "%s Duplicating page translations",
        logging_prefix + ("└─" if source_page.is_leaf() else "├─"),
    )
    # Clone all page translations of the source page
    source_page_translations = source_page.translations.all()
    num_translations = len(source_page_translations)
    translation_row_logging_prefix = logging_prefix + (
        "  " if source_page.is_leaf() else "│  "
    )
    for i, page_translation in enumerate(source_page_translations):
        # Set the page of the source translation to the new page
        page_translation.page = target_page
        # Delete the primary key to duplicate the object instance instead of updating it
        page_translation.pk = None
        # Set the translation to draft
        page_translation.status = status.DRAFT
        # Check if the page translation is valid
        page_translation.full_clean()
        # Save duplicated page translation
        page_translation.save(update_timestamp=False)
        logger.debug(
            "%s Created %r",
            translation_row_logging_prefix
            + ("└─" if i == num_translations - 1 else "├─"),
            page_translation,
        )


def duplicate_imprint(source_region, target_region):
    """
    Function to duplicate the imprint from one region to another.

    :param source_region: the source region from which the imprint should be duplicated
    :type source_region: ~integreat_cms.cms.models.regions.region.Region

    :param target_region: the target region
    :type target_region: ~integreat_cms.cms.models.regions.region.Region

    """
    source_imprint = source_region.imprint
    target_imprint = deepcopy(source_imprint)
    target_imprint.region = target_region
    # Delete the primary key to duplicate the object instance instead of updating it
    target_imprint.pk = None
    # Check if the new imprint object is valid
    target_imprint.full_clean()

    target_imprint.save()

    # Duplicate imprint translations by iterating to all existing ones
    source_page_translations = source_imprint.translations.all()

    for imprint_translation in source_page_translations:
        # Set the page of the source translation to the new imprint
        imprint_translation.page = target_imprint
        # Delete the primary key to duplicate the object instance instead of updating it
        imprint_translation.pk = None
        # Set the translation to draft
        imprint_translation.status = status.DRAFT
        # Check if the imprint translation is valid
        imprint_translation.full_clean()
        # Save duplicated imprint translation
        imprint_translation.save(update_timestamp=False)


# pylint: disable=unused-argument
def duplicate_media(source_region, target_region):
    """
    Function to duplicate all media of one region to another.

    :param source_region: the source region from which the pages should be duplicated
    :type source_region: ~integreat_cms.cms.models.regions.region.Region

    :param target_region: the target region
    :type target_region: ~integreat_cms.cms.models.regions.region.Region

    """
    # TODO: implement duplication of all media files



from django.utils.translation import ugettext_lazy as _

LINKCHECK_STATUS_TRANSLATIONS = [
    _("Broken external hash anchor"),
    _("Broken internal link"),
    _("Broken internal hash anchor"),
    _("Broken redirect"),
    _("Email link (not automatically checked)"),
    _("Empty link"),
    _("Failed to parse HTML for anchor"),
    _("Invalid URL"),
    _("Link to within the same page (not automatically checked)"),
    _("Missing Document"),
    _("Other Error: The read operation timed out"),
    _("Page OK but anchor can't be checked"),
    _("Phone number (not automatically checked)"),
    _("URL Not Yet Checked"),
    _("Working file link"),
    _("Working internal hash anchor"),
    _("Working internal link"),
    _("Working redirect"),
]



"""
This module contains the API client to interact with the SUMM.AI API
"""
import asyncio
import logging
import itertools

import aiohttp

from django.conf import settings

from .utils import TranslationHelper

logger = logging.getLogger(__name__)


class SummAiException(Exception):
    """
    Custom Exception class for errors during interaction with SUMM.AI
    """


class SummAiApiClient:
    """
    SUMM.AI API client to get German pages in Easy German language.
    """

    #: The current request
    #:
    #: :type: ~django.http.HttpRequest
    request = None
    #: The current region
    #:
    #: :type: ~integreat_cms.cms.models.regions.region.Region
    region = None
    #: The :class:`~integreat_cms.cms.forms.custom_content_model_form.CustomContentModelForm`
    #: subclass of the current content type
    #:
    #: :type: ~django.forms.models.ModelFormMetaclass
    form_class = None

    def __init__(self, request, form_class):
        """
        Constructor initializes the class variables

        :param region: The current region
        :type region: ~integreat_cms.cms.models.regions.region.Region

        :param form_class: The :class:`~integreat_cms.cms.forms.custom_content_model_form.CustomContentModelForm`
                           subclass of the current content type
        :type form_class: ~django.forms.models.ModelFormMetaclass
        """
        self.request = request
        self.region = request.region
        self.form_class = form_class

    async def translate_text_field(self, session, text_field):
        """
        Uses :meth:`aiohttp.ClientSession.post` to perform an asynchronous POST request to the SUMM.AI API.
        After the translation is finished, the processing is delegated to the specific textfield's
        :meth:`~integreat_cms.summ_ai_api.utils.TextField.translate`.

        :param session: The session object which is used for the request
        :type session: aiohttp.ClientSession

        :param text_field: The text field to be translated
        :type text_field: ~integreat_cms.summ_ai_api.utils.TextField

        :return: The modified text field containing the translated text
        :rtype: ~integreat_cms.summ_ai_api.utils.TextField
        """
        logger.debug("Translating %r", text_field)
        # Use test region for development
        user = settings.TEST_REGION_SLUG if settings.DEBUG else self.region.slug
        try:
            async with session.post(
                settings.SUMM_AI_API_URL,
                headers={"Authorization": f"Bearer {settings.SUMM_AI_API_KEY}"},
                json={
                    "input_text": text_field.text,
                    "user": user,
                    "separator": settings.SUMM_AI_SEPARATOR,
                    "is_test": settings.SUMM_AI_TEST_MODE,
                },
            ) as response:
                # Wait for the response
                response_data = await response.json()
                # Check whether the text was translated successfully
                if "translated_text" not in response_data:
                    if "error" in response_data:
                        raise SummAiException(
                            f"API error: {response.status} - {response_data['error']}"
                        )
                    raise SummAiException(
                        f"Unexpected API result: {response.status} - {response_data!r}"
                    )
                # Let the field handle the translated text
                text_field.translate(response_data["translated_text"])
                return text_field
        except (aiohttp.ClientError, SummAiException) as e:
            logger.error(
                "SUMM.AI translation of %r failed because of %s: %s",
                text_field,
                type(e),
                e,
            )
            text_field.exception = e

    async def translate_text_fields(self, loop, text_fields):
        """
        Translate a list of text fields from German into Easy German.
        Create an async task
        :meth:`~integreat_cms.summ_ai_api.summ_ai_api_client.SummAiApiClient.translate_text_field`
        for each entry.

        :param loop: The asyncio event loop
        :type loop: asyncio.AbstractEventLoop

        :param text_fields: The text fields to be translated
        :type text_fields: list [ ~integreat_cms.summ_ai_api.utils.TextField ]

        :returns: The list of completed text fields
        :rtype: list [ ~integreat_cms.summ_ai_api.utils.TextField ]
        """
        async with aiohttp.ClientSession() as session:
            # Create tasks for each text field
            tasks = [
                loop.create_task(self.translate_text_field(session, text_field))
                for text_field in text_fields
            ]
            # Wait for all tasks to finish and collect the results
            # (the results are sorted in the order the tasks were created)
            return await asyncio.gather(*tasks)

    def translate_queryset(self, queryset):
        """
        Translate a queryset of content objects from German into Easy German.

        To increase the speed of the translations, all operations are parallelized.

        :param queryset: The queryset which should be translated
        :type queryset: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel ]
        """
        # Make sure both languages exist
        self.request.region.get_language_or_404(settings.SUMM_AI_GERMAN_LANGUAGE_SLUG)
        easy_german = self.request.region.get_language_or_404(
            settings.SUMM_AI_EASY_GERMAN_LANGUAGE_SLUG
        )

        # Initialize translation helpers for each object instance
        translation_helpers = [
            TranslationHelper(self.request, self.form_class, object_instance)
            for object_instance in queryset
        ]

        # Aggregate all strings that need to be translated
        text_fields = itertools.chain(
            *[
                translation_helper.get_text_fields()
                for translation_helper in translation_helpers
            ]
        )

        # Initialize async event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # Translate queryset asynchronously in parallel
        loop.run_until_complete(self.translate_text_fields(loop, text_fields))

        # Commit changes to the database
        for translation_helper in translation_helpers:
            translation_helper.commit(easy_german)



import logging
import json

from django.conf import settings
from django.http import JsonResponse
from django.views.decorators.http import require_POST

from ...constants import status
from ...utils.user_utils import search_users
from ...models import (
    Region,
    EventTranslation,
    POITranslation,
    Feedback,
    PushNotificationTranslation,
    MediaFile,
    Directory,
)

logger = logging.getLogger(__name__)

# The maximum number of results returned by `search_content_ajax`
MAX_RESULT_COUNT = 20


def format_object_translation(object_translation, typ):
    """
    Formats the [poi/event/page]-translation as json

    :param object_translation: A translation object which has a title and a permalink
    :type object_translation: ~integreat_cms.cms.models.events.event.Event or ~integreat_cms.cms.models.pages.page.Page or ~integreat_cms.cms.models.pois.poi.POI
    :param typ: The type of this object
    :type typ: str
    :return: A dictionary with the title, url and type of the translation object
    :rtype: dict
    """
    return {
        "title": object_translation.title,
        "url": f"{settings.WEBAPP_URL}{object_translation.get_absolute_url()}",
        "type": typ,
    }


@require_POST
# pylint: disable=unused-argument
def search_content_ajax(request, region_slug=None, language_slug=None):
    """Searches all pois, events and pages for the current region and returns all that
    match the search query. Results which match the query in the title or slug get ranked
    higher than results which only match through their text content.

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param region_slug: The slug of the current region
    :type region_slug: str

    :param language_slug: language slug
    :type language_slug: str

    :raises AttributeError: If the request contains an object type which is unknown or if the user has no permission for it

    :return: Json object containing all matching elements, of shape {title: str, url: str, type: str}
    :rtype: ~django.http.JsonResponse
    """

    region = request.region
    body = json.loads(request.body.decode("utf-8"))
    query = body["query_string"]
    # whether to return only archived object, ignored if not applicable
    archived_flag = body["archived"]
    object_types = set(body.get("object_types", []))

    logger.debug("Ajax call: Live search for %r with query %r", object_types, query)

    results = []

    user = request.user
    if user.has_perm("cms.view_event") and "event" in object_types:
        object_types.remove("event")
        event_translations = (
            EventTranslation.search(region, language_slug, query)
            .filter(event__archived=archived_flag, status=status.PUBLIC)
            .select_related("event__region", "language")
        )
        results.extend(
            format_object_translation(obj, "event") for obj in event_translations
        )

    if user.has_perm("cms.view_feedback") and "feedback" in object_types:
        object_types.remove("feedback")
        results.extend(
            {
                "title": feedback.comment,
                "url": None,
                "type": "feedback",
            }
            for feedback in Feedback.search(region, query)
        )

    if user.has_perm("cms.view_page") and "page" in object_types:
        object_types.remove("page")
        pages = region.pages.all().cache_tree(archived=archived_flag)
        for page in pages:
            page_translation = page.get_translation(language_slug)
            if page_translation and (
                query.lower() in page_translation.slug
                or query.lower() in page_translation.title.lower()
            ):
                results.append(format_object_translation(page_translation, "page"))

    if user.has_perm("cms.view_poi") and "poi" in object_types:
        object_types.remove("poi")
        poi_translations = (
            POITranslation.search(region, language_slug, query)
            .filter(poi__archived=archived_flag, status=status.PUBLIC)
            .select_related("poi__region", "language")
        )
        results.extend(
            format_object_translation(obj, "poi") for obj in poi_translations
        )

    if (
        user.has_perm("cms.view_pushnotification")
        and "push_notification" in object_types
    ):
        object_types.remove("push_notification")
        results.extend(
            {
                "title": push_notification.title,
                "url": None,
                "type": "push_notification",
            }
            for push_notification in PushNotificationTranslation.search(
                region, language_slug, query
            )
        )

    if user.has_perm("cms.view_region") and "region" in object_types:
        object_types.remove("region")
        results.extend(
            {
                "title": region.name,
                "url": None,
                "type": "region",
            }
            for region in Region.search(query)
        )

    if user.has_perm("cms.view_user") and "user" in object_types:
        object_types.remove("user")
        results.extend(
            {
                "title": user.username,
                "url": None,
                "type": "user",
            }
            for user in search_users(region, query)
        )

    if "media" in object_types:
        object_types.remove("media")
        results.extend(
            {
                "title": file.name,
                "url": None,
                "type": "file",
            }
            for file in MediaFile.search(region, query)
        )
        results.extend(
            {
                "title": directory.name,
                "url": None,
                "type": "directory",
            }
            for directory in Directory.search(region, query)
        )

    if object_types:
        raise AttributeError(f"Unexpected object type(s): {object_types}")

    # sort alphabetically by title
    results.sort(key=lambda k: k["title"])

    return JsonResponse({"data": results[:MAX_RESULT_COUNT]})



import logging

from django import forms
from django.utils.translation import ugettext_lazy as _

from ..custom_filter_form import CustomFilterForm
from ...constants import translation_status

logger = logging.getLogger(__name__)


class PageFilterForm(CustomFilterForm):
    """
    Form for filtering page objects
    """

    translation_status = forms.MultipleChoiceField(
        label=_("Translation status"),
        widget=forms.CheckboxSelectMultiple(attrs={"class": "default-checked"}),
        choices=translation_status.CHOICES,
        initial=[key for (key, val) in translation_status.CHOICES],
        required=False,
    )
    query = forms.CharField(required=False)

    def __init__(self, **kwargs):
        r"""
        Initialize page filter form

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict
        """
        # Instantiate Form
        super().__init__(**kwargs)
        logger.debug("PageFilterForm initialized with data %r", self.data)

    def apply(self, pages, language_slug):
        """
        Filter the pages list according to the given filter data

        :param pages: The list of pages
        :type pages: list

        :param language_slug: The slug of the current language
        :type language_slug: str

        :return: The filtered page list
        :rtype: list
        """
        if self.is_enabled:
            logger.debug("Page tree filtered with changed data %r", self.changed_data)
            if "query" in self.changed_data:
                pages = self.filter_by_query(pages, language_slug)
            if "translation_status" in self.changed_data:
                pages = self.filter_by_translation_status(pages, language_slug)
        return pages

    def filter_by_query(self, pages, language_slug):
        """
        Filter the pages list by a given search query

        :param pages: The list of pages
        :type pages: list

        :param language_slug: The slug of the current language
        :type language_slug: str

        :return: The filtered page list
        :rtype: list
        """
        query = self.cleaned_data["query"].lower()
        # Buffer variable because the pages list should not be modified during iteration
        filtered_pages = []
        for page in pages:
            translation = page.get_translation(language_slug)
            if translation and (
                query in translation.slug or query in translation.title.lower()
            ):
                filtered_pages.append(page)
        return filtered_pages

    def filter_by_translation_status(self, pages, language_slug):
        """
        Filter the pages list by a given translation status

        :param pages: The list of pages
        :type pages: list

        :param language_slug: The slug of the current language
        :type language_slug: str

        :return: The filtered page list
        :rtype: list
        """
        selected_status = self.cleaned_data["translation_status"]
        # Buffer variable because the pages list should not be modified during iteration
        filtered_pages = []
        for page in pages:
            _, translation_state = page.translation_states.get(language_slug)
            if translation_state in selected_status:
                filtered_pages.append(page)
        return filtered_pages
